/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * CRWFileReader - implementation of RAW reader: Canon CRW specific parts.
 * It is heavily based on converter by Dave Coffin <dcoffin@shore.net>
******************************************************************************/

#include "stdafx.h"

#include "CRWFileReader.h"
#include "CRWImageInfo.h"
#include "Image.h"
#include "ImageUtilities.h"
//-----------------------------------------------------------------------------
const ulong CRWFileReader::m_buffer_size = 32768;
//-----------------------------------------------------------------------------
const uchar CRWFileReader::m_1st_tree[3][29] =
  {
    { 0,1,4,2,3,1,2,0,0,0,0,0,0,0,0,0,
      0x04,0x03,0x05,0x06,0x02,0x07,0x01,0x08,0x09,0x00,0x0a,0x0b,0xff  },

    { 0,2,2,3,1,1,1,1,2,0,0,0,0,0,0,0,
      0x03,0x02,0x04,0x01,0x05,0x00,0x06,0x07,0x09,0x08,0x0a,0x0b,0xff  },

    { 0,0,6,3,1,1,2,0,0,0,0,0,0,0,0,0,
      0x06,0x05,0x07,0x04,0x08,0x03,0x09,0x02,0x00,0x0a,0x01,0x0b,0xff  },
  };

const uchar CRWFileReader::m_2nd_tree[3][180] =
  {
    { 0,2,2,2,1,4,2,1,2,5,1,1,0,0,0,139,
      0x03,0x04,0x02,0x05,0x01,0x06,0x07,0x08,
      0x12,0x13,0x11,0x14,0x09,0x15,0x22,0x00,0x21,0x16,0x0a,0xf0,
      0x23,0x17,0x24,0x31,0x32,0x18,0x19,0x33,0x25,0x41,0x34,0x42,
      0x35,0x51,0x36,0x37,0x38,0x29,0x79,0x26,0x1a,0x39,0x56,0x57,
      0x28,0x27,0x52,0x55,0x58,0x43,0x76,0x59,0x77,0x54,0x61,0xf9,
      0x71,0x78,0x75,0x96,0x97,0x49,0xb7,0x53,0xd7,0x74,0xb6,0x98,
      0x47,0x48,0x95,0x69,0x99,0x91,0xfa,0xb8,0x68,0xb5,0xb9,0xd6,
      0xf7,0xd8,0x67,0x46,0x45,0x94,0x89,0xf8,0x81,0xd5,0xf6,0xb4,
      0x88,0xb1,0x2a,0x44,0x72,0xd9,0x87,0x66,0xd4,0xf5,0x3a,0xa7,
      0x73,0xa9,0xa8,0x86,0x62,0xc7,0x65,0xc8,0xc9,0xa1,0xf4,0xd1,
      0xe9,0x5a,0x92,0x85,0xa6,0xe7,0x93,0xe8,0xc1,0xc6,0x7a,0x64,
      0xe1,0x4a,0x6a,0xe6,0xb3,0xf1,0xd3,0xa5,0x8a,0xb2,0x9a,0xba,
      0x84,0xa4,0x63,0xe5,0xc5,0xf3,0xd2,0xc4,0x82,0xaa,0xda,0xe4,
      0xf2,0xca,0x83,0xa3,0xa2,0xc3,0xea,0xc2,0xe2,0xe3,0xff,0xff  },

    { 0,2,2,1,4,1,4,1,3,3,1,0,0,0,0,140,
      0x02,0x03,0x01,0x04,0x05,0x12,0x11,0x06,
      0x13,0x07,0x08,0x14,0x22,0x09,0x21,0x00,0x23,0x15,0x31,0x32,
      0x0a,0x16,0xf0,0x24,0x33,0x41,0x42,0x19,0x17,0x25,0x18,0x51,
      0x34,0x43,0x52,0x29,0x35,0x61,0x39,0x71,0x62,0x36,0x53,0x26,
      0x38,0x1a,0x37,0x81,0x27,0x91,0x79,0x55,0x45,0x28,0x72,0x59,
      0xa1,0xb1,0x44,0x69,0x54,0x58,0xd1,0xfa,0x57,0xe1,0xf1,0xb9,
      0x49,0x47,0x63,0x6a,0xf9,0x56,0x46,0xa8,0x2a,0x4a,0x78,0x99,
      0x3a,0x75,0x74,0x86,0x65,0xc1,0x76,0xb6,0x96,0xd6,0x89,0x85,
      0xc9,0xf5,0x95,0xb4,0xc7,0xf7,0x8a,0x97,0xb8,0x73,0xb7,0xd8,
      0xd9,0x87,0xa7,0x7a,0x48,0x82,0x84,0xea,0xf4,0xa6,0xc5,0x5a,
      0x94,0xa4,0xc6,0x92,0xc3,0x68,0xb5,0xc8,0xe4,0xe5,0xe6,0xe9,
      0xa2,0xa3,0xe3,0xc2,0x66,0x67,0x93,0xaa,0xd4,0xd5,0xe7,0xf8,
      0x88,0x9a,0xd7,0x77,0xc4,0x64,0xe2,0x98,0xa5,0xca,0xda,0xe8,
      0xf3,0xf6,0xa9,0xb2,0xb3,0xf2,0xd2,0x83,0xba,0xd3,0xff,0xff  },

    { 0,0,6,2,1,3,3,2,5,1,2,2,8,10,0,117,
      0x04,0x05,0x03,0x06,0x02,0x07,0x01,0x08,
      0x09,0x12,0x13,0x14,0x11,0x15,0x0a,0x16,0x17,0xf0,0x00,0x22,
      0x21,0x18,0x23,0x19,0x24,0x32,0x31,0x25,0x33,0x38,0x37,0x34,
      0x35,0x36,0x39,0x79,0x57,0x58,0x59,0x28,0x56,0x78,0x27,0x41,
      0x29,0x77,0x26,0x42,0x76,0x99,0x1a,0x55,0x98,0x97,0xf9,0x48,
      0x54,0x96,0x89,0x47,0xb7,0x49,0xfa,0x75,0x68,0xb6,0x67,0x69,
      0xb9,0xb8,0xd8,0x52,0xd7,0x88,0xb5,0x74,0x51,0x46,0xd9,0xf8,
      0x3a,0xd6,0x87,0x45,0x7a,0x95,0xd5,0xf6,0x86,0xb4,0xa9,0x94,
      0x53,0x2a,0xa8,0x43,0xf5,0xf7,0xd4,0x66,0xa7,0x5a,0x44,0x8a,
      0xc9,0xe8,0xc8,0xe7,0x9a,0x6a,0x73,0x4a,0x61,0xc7,0xf4,0xc6,
      0x65,0xe9,0x72,0xe6,0x71,0x91,0x93,0xa6,0xda,0x92,0x85,0x62,
      0xf3,0xc5,0xb2,0xa4,0x84,0xba,0x64,0xa5,0xb3,0xd2,0x81,0xe5,
      0xd3,0xaa,0xc4,0xca,0xf2,0xb1,0xe4,0xd1,0x83,0x63,0xea,0xc3,
      0xe2,0x82,0xf1,0xa3,0xc2,0xa1,0xc1,0xe3,0xa2,0xe1,0xff,0xff  }
  };
//-----------------------------------------------------------------------------
CRWFileReader::CRWFileReader(): IParametrizable("CRWFileReader"), mp_image_info(NULL)
  {
  }
//-----------------------------------------------------------------------------
CRWFileReader::~CRWFileReader()
  {
  assert(mp_image_info == NULL);
  delete mp_image_info;
  }
//-----------------------------------------------------------------------------
RawFileReader& CRWFileReader::GetRawReader()
  {
  static CRWFileReader reader;
  return reader;
  }
//-----------------------------------------------------------------------------
void CRWFileReader::ParseCiffTag(ushort type, ulong len, ulong abs_offset)
  {
  switch (type)
    {
    case 0x1835:
      mg_log << TRACE << "Canon decompress tables CIFF tag detected" << endlog;
      if (!m_fl_parse_only)
        {
        Seek(abs_offset, std::ios::beg);
        mg_log << TRACE << "Canon decompress tables init" << endlog;
        InitTables(Get4());
        }
      break;

    case 0x2005:
      mg_log << TRACE << "Canon CRW data detected, len=" << len << endlog;
      mp_image_info -> m_end_data_pos = abs_offset + len;
      break;

    case 0x2007:
      mg_log << TRACE << "Canon JPEG type 1 thumbnail detected, len=" << len << endlog;

      if (!m_fl_small_icon_mode)
        {
        mp_image_info -> m_thumbnails.push_back(CRWImageInfo::Thumbnail(len));
        Seek(abs_offset, std::ios::beg);
        m_file.read(&mp_image_info -> m_thumbnails.back()[0], len);
        }
      break;

    case 0x2008:
      mg_log << TRACE << "Canon JPEG type 2 thumbnail detected, len=" << len << endlog;
      mp_image_info -> m_thumbnails.push_back(CRWImageInfo::Thumbnail(len));
      Seek(abs_offset, std::ios::beg);
      m_file.read(&mp_image_info -> m_thumbnails.back()[0], len);
      break;

    case 0x102a:
      mg_log << TRACE << "Canon shot info CIFF tag detected" << endlog;
      Seek(abs_offset + 0x0E, std::ios::beg);
      mp_image_info -> m_light_src = m_file.get();
      mg_log << INFO << "White balance setting: " << mp_image_info -> m_light_src_descr[mp_image_info -> m_light_src] << endlog;
      break;

    case 0x102c:
      {
      if (mp_image_info -> m_model_name != "Canon PowerShot S30" &&
          mp_image_info -> m_model_name != "Canon PowerShot S40" &&
          mp_image_info -> m_model_name != "Canon PowerShot G2")
        break;

      mg_log << TRACE << "Canon S40/G2 white balance CIFF tag detected" << endlog;
      Seek(abs_offset + 0x64, std::ios::beg);

      ushort green1 = Get2(),
             red = Get2(),
             blue = Get2(),
             green2 = Get2();

      mp_image_info -> m_red_wb = (float)red/green1;
      mp_image_info -> m_blue_wb = (float)blue/green2;

      mg_log << INFO << "White balance coefficients: "
        << mp_image_info -> m_red_wb << ", " << mp_image_info -> m_blue_wb << endlog;
      break;
      }

    case 0x32:
      {
      if (mp_image_info -> m_model_name != "Canon EOS D30" ||
          !mp_image_info -> m_light_src)
        break;

      mg_log << TRACE << "Canon EOS D30 white balance CIFF tag detected" << endlog;

      Seek(abs_offset + 72, std::ios::beg);

      ushort red = Get2(),
             green1 = Get2(),
             blue = Get2(),
             green2 = Get2();

      mp_image_info -> m_red_wb = (float)red/green1;
      mp_image_info -> m_blue_wb = (float)blue/green2;

      mg_log << INFO << "White balance coefficients: "
        << mp_image_info -> m_red_wb << ", " << mp_image_info -> m_blue_wb << endlog;
      }
      break;

    case 0x10A9:
      {
      if (mp_image_info -> m_model_name != "Canon EOS D60")
        break;

      mg_log << TRACE << "Canon EOS D60 white balance CIFF tag detected" << endlog;

      Seek(abs_offset + 2 + 8*mp_image_info -> m_light_src, std::ios::beg);

      ushort red = Get2(),
             green1 = Get2(),
             green2 = Get2(),
             blue = Get2();

      mp_image_info -> m_red_wb = (float)red/green1;
      mp_image_info -> m_blue_wb = (float)blue/green2;

      mg_log << INFO << "White balance coefficients: "
        << mp_image_info -> m_red_wb << ", " << mp_image_info -> m_blue_wb << endlog;
      }
      break;
    }
  }
//-----------------------------------------------------------------------------
void CRWFileReader::InitTables(ulong table)
  {
  assert(table <= 2);
  if (table > 2) table = 2;
  memset(m_1st_decode, 0, sizeof m_1st_decode);
  memset(m_2nd_decode, 0, sizeof m_2nd_decode);
  MakeDecoder(m_1st_decode,  m_1st_tree[table], 0);
  MakeDecoder(m_2nd_decode, m_2nd_tree[table], 0);
  }
//-----------------------------------------------------------------------------
void CRWFileReader::MakeDecoder(Decode *dest, const uchar *source, unsigned level)
  {
  if (!level)
    {
    mp_free = dest;
    m_leaf = 0;
    }

  mp_free++;

 // At what level should the next leaf appear?

  unsigned i, next;
  for (i = next = 0; i <= m_leaf && next < 16; )
    i += source[next++];

  if (level < next) // Are we there yet?
    {
    dest -> mp_branch[0] = mp_free;
    MakeDecoder(mp_free, source, level + 1);
    dest-> mp_branch[1] = mp_free;
    MakeDecoder(mp_free, source, level + 1);
    }
    else
      dest -> m_leaf = source[16 + m_leaf++];
  }
//-----------------------------------------------------------------------------
inline void CRWFileReader::GetNextBlock()
  {
  ulong prev_pos = m_file.tellg();

  if (m_eof)
    {
    mg_log << CRIT << "CRW data error at " << prev_pos << ": attempt to read beyond the end of CRW data block" << endlog;
    throw ExRawFileError("RawDataError", "CRW data error");
    }

  if ((ulong)m_file.tellg() + m_buffer_size > mp_image_info -> m_end_data_pos)
    {
    m_curr_buffer_size = mp_image_info -> m_end_data_pos - m_file.tellg();

    if (m_curr_buffer_size > 0)
      m_file.read(&m_buffer[0], m_curr_buffer_size);
    else
      m_eof = true;
    }
  else
    {
    m_curr_buffer_size = m_buffer_size;
    m_file.read(&m_buffer[0], m_buffer_size);
    }

  if ((ulong)m_file.tellg() - prev_pos != m_curr_buffer_size)
    {
    mg_log << CRIT << "CRW data error at " << (ulong)m_file.tellg() - m_curr_buffer_size + m_buffer_pos << " : got block of size " << (ulong)m_file.tellg() - prev_pos << " instead of " << m_curr_buffer_size << endlog;
    throw ExRawFileError("RawDataError", "CRW data error");
    }

  m_buffer_pos = 0;
  }
//-----------------------------------------------------------------------------
//  GetBits(n) where 0 <= n <= 25 returns an n-bit integer
//
inline ulong CRWFileReader::GetBits(short nbits)
  {
  if (nbits == 0) return 0;

  if (nbits == -1)
    {
    m_ret = 0;
    m_bitbuf = 0;
    m_vbits = 0;
    }
  else
    {
    m_ret = m_bitbuf << (32 - m_vbits) >> (32 - nbits);

    if (m_vbits < nbits)
      {
      mg_log << CRIT << "CRW data error at " << (ulong)m_file.tellg() - m_curr_buffer_size + m_buffer_pos << ": attempt to read more bits that were cached" << endlog;
      throw ExRawFileError("RawDataError", "CRW data error");
      }

    m_vbits -= nbits;
    }

  while (m_vbits < 25 && !m_eof)
    {
    uchar c = m_buffer[m_buffer_pos++];

    m_bitbuf = (m_bitbuf << 8) + c;

    if (m_buffer_pos >= m_curr_buffer_size)
      GetNextBlock();

    // Canon puts an extra 0 after 0xff
    if (c == 0xff)
      {
      // It seems in some CRW files in the very end of data block
      // after 0xFF there is not 0x00 but 0xD9,
      // thus we use modified this check in the very end of data block
      if (m_buffer[m_buffer_pos] != 0 &&
          ((ulong)m_file.tellg() - m_curr_buffer_size + m_buffer_pos < mp_image_info -> m_end_data_pos - 100
           || (uchar)m_buffer[m_buffer_pos] != 0xD9))
        {
        mg_log << WARN << "CRW data error at " << (ulong)m_file.tellg() - m_curr_buffer_size + m_buffer_pos << ": 0x" << std::hex << (int)(uchar)m_buffer[m_buffer_pos] << std::dec << " was detected after 0xFF instead of 0x00" << endlog;
        mg_log << WARN << "CRW file is possibly corrupted" << endlog;
        //throw ExRawFileError("RawDataError", "CRW data error");
        }

      ++m_buffer_pos;

      if (m_buffer_pos >= m_curr_buffer_size)
        GetNextBlock();
      }

    m_vbits += 8;
    }

  return m_ret;
  }
//-----------------------------------------------------------------------------
void CRWFileReader::DecompressInit(ulong count)
  {
  m_carry = m_pixel = 0;
  m_buffer.resize(m_buffer_size);
  Seek(count, std::ios::beg);

  m_file.read(&m_buffer[0], m_buffer_size);
  m_curr_buffer_size = m_buffer_size;
  m_buffer_pos = 0;
  m_eof = false;

  GetBits(-1);
  }
//-----------------------------------------------------------------------------
// Decompress "count" blocks of 64 samples each
//
void CRWFileReader::Decompress(std::vector<ushort>& outbuf, unsigned count)
  {
  Decode *decode, *dindex;
  long diffbuf[64];
  long outpos = 0;

  while (count--)
    {
    memset(diffbuf, 0, sizeof diffbuf);

    decode = m_1st_decode;

    for (ushort i = 0; i < 64; ++i)
      {
      for (dindex = decode; dindex -> mp_branch[0]; )
	      dindex = dindex -> mp_branch[GetBits(1)];

      unsigned leaf = dindex -> m_leaf;
      decode = m_2nd_decode;

      if (leaf == 0 && i) break;
      if (leaf == 0xff) continue;

      i += leaf >> 4;
      unsigned len = leaf & 15;
      if (len == 0) continue;

      uchar sign = (GetBits(1)); // 1 is positive, 0 is negative
      long diff = GetBits(len - 1);

      if (sign)
	      diff += 1 << (len-1);
      else
	      diff += (-1 << len) + 1;

      if (i < 64) diffbuf[i] = diff;
      }

    diffbuf[0] += m_carry;
    m_carry = diffbuf[0];

    for (ushort i = 0; i < 64; ++i)
      {
      if (m_pixel++ % mp_image_info -> GetWidth() == 0)
	      m_base[0] = m_base[1] = 512;

      outbuf[outpos++] = (m_base[i & 1] += diffbuf[i]);
      }
    }
  }
//-----------------------------------------------------------------------------
void CRWFileReader::Load(Image& image)
  {
  m_fl_12bits = false;
  ushort clamp_val;
  unsigned width, height, left, top;

  // 10 01 01 00
  mp_image_info -> m_filters = 0x94;

  if (mp_image_info -> m_model_name == "Canon PowerShot S40" ||
      mp_image_info -> m_model_name == "Canon PowerShot G2")
    {
    clamp_val = 0xFFC0;
    width = 2312;
    height = 1720;

    left = 12;
    top = 6;
    }
  else if (mp_image_info -> m_model_name == "Canon PowerShot S30")
    {
    clamp_val = 0xFFC0;
    width  = 2088;
    height = 1550;
    left = 4;
    top = 8;
    // Don't have S30 RAW file to experiment
    if (!mp_image_info -> m_red_wb)
      {
      mp_image_info -> m_red_wb = 1.785;
      mp_image_info -> m_blue_wb = 1.266;
      }
    }
  else if (mp_image_info -> m_model_name == "Canon PowerShot S45" ||
           mp_image_info -> m_model_name == "Canon PowerShot G3")
    {
    m_fl_12bits = true;
    clamp_val = 0xFFF0;
    width = 2312;
    height = 1720;
    left = 12;
    top = 6;
    // Currently we cannot determine WB settings for these cameras :-(
    if (!mp_image_info -> m_red_wb)
      {
      mp_image_info -> m_red_wb = 1.966;
      mp_image_info -> m_blue_wb = 1.208;
      }
    }
  else if (mp_image_info -> m_model_name == "Canon EOS D30")
    {
    clamp_val = 0xFFF0;
    m_fl_12bits = true;
    left = 48;
    top = 6;
    width = 2176;
    height = 1448;
    if (!mp_image_info -> m_red_wb)
      {
      mp_image_info -> m_red_wb = 1.592;
      mp_image_info -> m_blue_wb = 1.261;
      }
    }
  else if (mp_image_info -> m_model_name == "Canon EOS D60")
    {
    clamp_val = 0xFFF0;
    m_fl_12bits = true;
    left = 64;
    top = 12;
    width = 3088;
    height = 2056;

    if (!mp_image_info -> m_red_wb)
      {
      mp_image_info -> m_red_wb = 2.242;
      mp_image_info -> m_blue_wb = 1.245;
      }
    }
  else if (mp_image_info -> m_model_name == "Canon EOS 10D")
    {
    clamp_val = 0xFFF0;
    m_fl_12bits = true;
    left = 64;
    top = 12;
    width = 3088;
    height = 2056;

    if (!mp_image_info -> m_red_wb)
      {
      mp_image_info -> m_red_wb = 2.242;
      mp_image_info -> m_blue_wb = 1.245;
      }
    }
  else
    {
    CleanUp();
    throw ExRawFileError("UnsupportedCamera", "Sorry, this camera is unsupported");
    }

  if (!m_fl_use_standard_size)
    {
    mp_image_info -> m_left_border = left;
    mp_image_info -> m_top_border = top;
    }
  else
    {
    width = mp_image_info -> m_std_width;
    height = mp_image_info -> m_std_height;
    }

  if (m_fl_leave_border)
    {
    mp_image_info -> m_left_border = 0;
    mp_image_info -> m_top_border = 0;
    width = mp_image_info -> m_width;
    height = mp_image_info -> m_height;
    }

  short angle = mp_image_info -> m_rotation_angle;
  if (angle < 0)
    angle += 360;

  if (angle == 90 || angle == 270)
    image.Create(height + 2*m_trim_width, width + 2*m_trim_width, clamp_val,
                 m_trim_width);
  else
    {
    image.Create(width + 2*m_trim_width, height + 2*m_trim_width, clamp_val,
                 m_trim_width);
    
    if (angle && angle != 180)
      {
      mg_log << WARN << "Angle is supposed to be 0, 90, 180 or 270, but angle " << angle << " is detected; assuming 0";
      angle = 0;
      }
    }

  std::vector<ushort> pixels(mp_image_info -> m_width * 8);

  mg_log << INFO << "Decompressing image from " << mp_image_info -> GetModelName() << " camera" << endlog; 

  if (m_fl_12bits)
    DecompressInit(540 + mp_image_info -> m_height*mp_image_info -> m_width/4);
  else
    DecompressInit(540);

  ulong64 noise[3] = {0, 0, 0};
  ulong noise_count[3] = {0, 0, 0};

  ushort mask;
  uchar shift;

  if (m_fl_12bits)
    {
    mask = 0xfff;
    shift = 4;
    }
  else
    {
    mask = 0x3ff;
    shift = 6;
    }

  mp_image_info -> m_max_color[0] = 0;
  mp_image_info -> m_max_color[1] = 0;
  mp_image_info -> m_max_color[2] = 0;

  image.Zero();

  uchar row_step = 4;

  if (!(mp_image_info -> m_height%8))
    row_step = 8;

  if (mp_image_info -> m_height%4 || (mp_image_info -> m_width*row_step)%64)
    {
    mg_log << CRIT << "Wrong image sizes - height%4 or width*row_step%64 are not zeroes" << endlog;
    mg_log << WARN << "CRW file is possibly corrupted or has unsupported type" << endlog;
    throw ExRawFileError("RawDataError", "CRW data error");
    }

  for (unsigned row = 0; row < mp_image_info -> m_height; row += row_step)
    {
    Decompress(pixels, mp_image_info -> m_width*row_step/64); // Get `row_step' rows

    if (m_fl_12bits) // Let's add low 2 bits
      {
      std::streampos save = m_file.tellg();

      Seek(26 + row * mp_image_info -> m_width/4, std::ios::beg);

      for (unsigned prow = 0, i = 0; i < mp_image_info -> m_width * row_step/4; ++i)
        {
        uchar c = m_file.get();

        for (uchar r = 0; r < 8; r += 2, ++prow)
	        pixels[prow] = (pixels[prow] << 2) + ((c >> r) & 3);
        }

      m_file.seekg(save);
      }

    for (uchar r = 0; r < row_step; ++r)
      for (unsigned col = 0; col < mp_image_info -> m_width; ++col)
        {
	    int irow = row + r - mp_image_info -> m_top_border;
	    int icol = col - mp_image_info -> m_left_border;
        uchar index = mp_image_info -> GetCurrentColorIndex(row + r, col);
        assert(index < 3);

        if (irow >= 0 && (unsigned)irow < height &&
            icol >= 0 && (unsigned)icol < width)
          {
          ushort color = (pixels[r * mp_image_info -> m_width + col] & mask) << shift;

          if (mp_image_info -> m_max_color[index] < color)
            mp_image_info -> m_max_color[index] = color;

          if (!angle)
            image.At(irow + m_trim_width,
                     icol + m_trim_width)[index] = color;
          else if (angle == 90)
            image.At(icol + m_trim_width,
                     height + m_trim_width - irow - 1)[index] = color;
          else if (angle == 180)
            image.At(height + m_trim_width - irow - 1,
                     icol + m_trim_width)[index] = color;
          else if (angle == 270)
            image.At(width + m_trim_width - icol - 1,
                     irow + m_trim_width)[index] = color;
          }
	      else if (row + r > 1 && col > 1 &&
                 row + r < mp_image_info -> m_height - 2 &&
                 col < mp_image_info -> m_width - 2 &&
                 !((int)(row + r) >= (int)mp_image_info -> m_top_border - 2 &&
                   (int)col >= (int)mp_image_info -> m_left_border - 2 &&
                   row + r <= mp_image_info -> m_height - 5 && col <= mp_image_info -> m_width - 5))
            {
            noise[index] += pixels[r * mp_image_info -> m_width + col] & mask;
            ++noise_count[index];
            }
        }
    }

  ulong file_pos = (ulong)m_file.tellg() - m_curr_buffer_size + m_buffer_pos;
  mg_log << TRACE << "File position: " << file_pos << endlog;

  if (file_pos != mp_image_info -> m_end_data_pos &&
      file_pos != mp_image_info -> m_end_data_pos - 1)
    {
    mg_log << CRIT << "File position after reading CRW data is " << file_pos
           << " instead of " << mp_image_info -> m_end_data_pos << endlog;

    throw ExRawFileError("RawDataError", "CRW data error");
    }
 
  mg_log << TRACE << "Max color[0]: " << mp_image_info -> m_max_color[0] << endlog;
  mg_log << TRACE << "Max color[1]: " << mp_image_info -> m_max_color[1] << endlog;
  mg_log << TRACE << "Max color[2]: " << mp_image_info -> m_max_color[2] << endlog;

  for (uchar i = 0; i < 3; ++i)
    {
    mg_log << DEBUG << "Noise samples number for channel " << (int)i << "=" << noise_count[i] << endlog;

    if (noise_count[i])
      mp_image_info -> m_noise[i] = ((ulong64) noise[i] << shift) / noise_count[i];
    else
      mp_image_info -> m_noise[i] = 0;
    }

  // Let's correct filters now:
  if (angle == 90)
    mp_image_info -> m_filters = 0x61; // 01 10 00 01
  else if (angle == 270 || angle == 180)
    mp_image_info -> m_filters = 0x49; // 01 00 10 01
  }
//-----------------------------------------------------------------------------
