/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef CRW_FILE_READER_H
#define CRW_FILE_READER_H

#include "RawFileReader.h"
#include "CRWImageInfo.h"
//-----------------------------------------------------------------------------
class CRWFileReader: public RawFileReader
  {
  public:
    static RawFileReader& GetRawReader();

    virtual RawImageInfo& GetImageInfo();

    RawImageInfo* RetrieveImageInfo();

  protected:
    CRWFileReader();
    ~CRWFileReader();

    virtual void ParseCiffTag(ushort type, ulong len, ulong abs_offset);

    virtual void Load(Image& p_image);

    virtual void PrepareImageInfo();

  private:
    CRWImageInfo* mp_image_info;
    struct Decode
      {
      Decode *mp_branch[2];
      unsigned m_leaf;
      } m_1st_decode[32], m_2nd_decode[512];

    static const uchar m_1st_tree[3][29];
    static const uchar m_2nd_tree[3][180];

    static const ulong m_buffer_size;

    Decode *mp_free;	// Next unused node
    unsigned m_leaf;         // Number of leaves already added
    std::vector<char> m_buffer;
    ulong m_buffer_pos, m_curr_buffer_size;
    
    ulong m_bitbuf, m_ret;
    int m_vbits;

    int m_carry, m_pixel, m_base[2];
 
    bool m_fl_12bits, m_eof;

  private:
    void InitTables(ulong table);

    void MakeDecoder(Decode *dest, const uchar *source, unsigned level);

    ulong GetBits(short num);

    uchar GetCurrentColorIndex(unsigned row, unsigned col) const;

    void DecompressInit(ulong count);

    void Decompress(std::vector<ushort>& outbuf, unsigned count);

    void GetNextBlock();
  };
//-----------------------------------------------------------------------------
inline RawImageInfo& CRWFileReader::GetImageInfo()
  {
  return *mp_image_info;
  }
//-----------------------------------------------------------------------------
inline void CRWFileReader::PrepareImageInfo()
  {
  assert (mp_image_info == NULL);

  if (!mp_image_info)
    mp_image_info = new CRWImageInfo;
  }
//-----------------------------------------------------------------------------
inline RawImageInfo* CRWFileReader::RetrieveImageInfo()
  {
  RawImageInfo* info = mp_image_info;
  mp_image_info = NULL;

  return info;
  }
//-----------------------------------------------------------------------------

#endif // CRW_FILE_READER_H
