/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * CornerInterpolator - experimental implementation of interpolation based on
 * ``corners shape detection'' idea. See copyrights.txt for copyright info.
******************************************************************************/

#include "stdafx.h"

#include "CornerInterpolator.h"
#include "Image.h"
#include "RawImageInfo.h"
//-----------------------------------------------------------------------------
CornerInterpolator::CornerInterpolator(): IParametrizable("CornerInterpolator")
  {
  m_small_threshold = 512;
  }
//-----------------------------------------------------------------------------
CornerInterpolator& CornerInterpolator::GetInterpolator()
  {
  static CornerInterpolator corner_int;
  return corner_int;
  }
//-----------------------------------------------------------------------------
void CornerInterpolator::GetParameters(Params& o_params) const
  {
  o_params["Corner_SmallThreshold"] = boost::lexical_cast<std::string>(m_small_threshold);
  }
//-----------------------------------------------------------------------------
void CornerInterpolator::GetDescriptions(Params& o_params) const
  {
  o_params["Corner_SmallThreshold"] = "If pixel value is smaller "
    "than this threshold value - safer interpolation method will be used for current pixel";
  }
//-----------------------------------------------------------------------------
void CornerInterpolator::SetParameters(Params& i_params)
  {
  if (i_params.exist("Corner_SmallThreshold"))
    m_small_threshold = boost::lexical_cast<ulong>(i_params.pop("Corner_SmallThreshold"));
  }
//-----------------------------------------------------------------------------
uchar CornerInterpolator::SplitColors(ColorIndex colors[], uchar num)
  {
  ushort diff = colors[1].m_color - colors[0].m_color;
  uchar ind = 0;

  for (uchar i = 1; i < num - 1; ++i)
    if (colors[i + 1].m_color - colors[i].m_color > diff)
      {
      diff = colors[i + 1].m_color - colors[i].m_color;
      ind = i;
      }

  return ind;
  }
//-----------------------------------------------------------------------------
void CornerInterpolator::InterpolateGreenPixel(Image& image, unsigned row, unsigned col)
  {
  uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
  assert (ind != 1);

  ColorIndex greens[4] = {ColorIndex(image.At(row - 1, col)[1], -1, 0),
			  ColorIndex(image.At(row + 1, col)[1], 1, 0),
			  ColorIndex(image.At(row, col - 1)[1], 0, -1),
			  ColorIndex(image.At(row, col + 1)[1], 0, 1)};

  std::sort(&greens[0], &greens[4]);

  ColorIndex chromas[5] = {ColorIndex(image.At(row, col)[ind], 0, 0),
			   ColorIndex(image.At(row - 2, col)[ind], -2, 0),
			   ColorIndex(image.At(row + 2, col)[ind], 2, 0),
			   ColorIndex(image.At(row, col - 2)[ind], 0, -2),
			   ColorIndex(image.At(row, col + 2)[ind], 0, 2)};

  std::sort(&chromas[0], &chromas[5]);

  uchar dark_green_ind = SplitColors(greens, 4);

  uchar dark_chroma_ind = SplitColors(chromas, 5);

  int av_green = 0;

  int dark_red_gr = 0;

  if (dark_green_ind == 0)
    dark_red_gr += (int)image.At(row, col)[ind]
		   - (int)image.At(row + 2*greens[0].m_row,
				   col + 2*greens[0].m_col)[ind];
  else if (dark_green_ind == 1)
    {
    for (uchar i = 0; i <= dark_green_ind; ++i)
      dark_red_gr += (int)image.At(row, col)[ind]
		     - (int)image.At(row + 2*greens[i].m_row,
				     col + 2*greens[i].m_col)[ind];

    dark_red_gr /= 2;
    }
  else if (dark_green_ind == 2)
    {
    if (greens[1].m_color - greens[0].m_color < greens[2].m_color - greens[1].m_color)
      {
      for (uchar i = 0; i <= 1; ++i)
	dark_red_gr += (int)image.At(row, col)[ind]
		       - (int)image.At(row + 2*greens[i].m_row,
				       col + 2*greens[i].m_col)[ind];

      dark_red_gr /= 2;
      }
    else
      {
      for (uchar i = 1; i <= 2; ++i)
	dark_red_gr += (int)image.At(row, col)[ind]
		       - (int)image.At(row + 2*greens[i].m_row,
				       col + 2*greens[i].m_col)[ind];

      dark_red_gr /= 2;
      }
    }

  int light_red_gr = 0;

  if (dark_green_ind == 2)
    light_red_gr += (int)image.At(row, col)[ind]
		    - (int)image.At(row + 2*greens[0].m_row,
				    col + 2*greens[0].m_col)[ind];
  else if (dark_green_ind == 1)
    {
    for (uchar i = dark_green_ind + 1; i <= 3; ++i)
      light_red_gr += (int)image.At(row, col)[ind]
		      - (int)image.At(row + 2*greens[i].m_row,
				      col + 2*greens[i].m_col)[ind];

    light_red_gr /= 2;
    }
  else if (dark_green_ind == 0)
    {
    if (greens[2].m_color - greens[1].m_color < greens[3].m_color - greens[2].m_color)
      {
      for (uchar i = 1; i <= 2; ++i)
	light_red_gr += (int)image.At(row, col)[ind]
			- (int)image.At(row + 2*greens[i].m_row,
					col + 2*greens[i].m_col)[ind];

      light_red_gr /= 2;
      }
    else
      {
      for (uchar i = 2; i <= 3; ++i)
	light_red_gr += (int)image.At(row, col)[ind]
			- (int)image.At(row + 2*greens[i].m_row,
					col + 2*greens[i].m_col)[ind];

      light_red_gr /= 2;
      }
    }

  uchar g_ind;

  bool fl_single = false;

  if (dark_red_gr < light_red_gr)
    {
    if (dark_green_ind == 1 || dark_green_ind > 1 && greens[1].m_color - greens[0].m_color < greens[2].m_color - greens[1].m_color)
      g_ind = 0;
    else if (dark_green_ind > 1)
      g_ind = 1;
    else
      {
      g_ind = 0;
      fl_single = true;
      }
    }
  else
    {
    if (dark_green_ind == 1 || dark_green_ind == 0 && greens[2].m_color - greens[1].m_color > greens[3].m_color - greens[2].m_color)
      g_ind = 2;
    else if (dark_green_ind == 0)
      g_ind = 1;
    else
      {
      g_ind = 3;
      fl_single = true;
      }
    }

  if (fl_single)
    {
    float chroma = ((int)image.At(row, col)[ind]
		    + (int)image.At(row + 2*greens[g_ind].m_row,
				    col + 2*greens[g_ind].m_col)[ind])/2.0;

    float hue = chroma/greens[g_ind].m_color;

    av_green = (int)(image.At(row, col)[ind]/hue);
    }
  else
    {
    float chroma0 = ((int)image.At(row, col)[ind]
		     + (int)image.At(row + 2*greens[g_ind].m_row,
				   col + 2*greens[g_ind].m_col)[ind])/2.0;

    float chroma1 = ((int)image.At(row, col)[ind]
		     + (int)image.At(row + 2*greens[g_ind+1].m_row,
				   col + 2*greens[g_ind+1].m_col)[ind])/2.0;

    float hue = (chroma0/greens[g_ind].m_color + chroma1/greens[g_ind+1].m_color)/2.0;

    av_green = (int)(image.At(row, col)[ind]/hue);
    }

  image.At(row, col)[1] = image.ClampGreenPixel(av_green, row, col);
  }
//-----------------------------------------------------------------------------
void CornerInterpolator::InterpolatePixelInCol(Image& image, unsigned row, unsigned col, uchar ind)
  {
  assert (mp_image_info -> GetCurrentColorIndex(row, col) == 1);

  int CC, P;

  CC = 2*(int)image.At(row, col)[1] - (int)image.At(row - 1, col)[1] - (int)image.At(row + 1, col)[1];
  if (image.At(row, col)[1] < m_small_threshold ||
      image.At(row - 1, col)[1] < m_small_threshold ||
      image.At(row + 1, col)[1] < m_small_threshold)
    P = ((int)image.At(row - 1, col)[ind] + (int)image.At(row + 1, col)[ind] + CC) / 2;
  else
    {
    float c0 = (float)image.At(row - 1, col)[ind]/image.At(row - 1, col)[1];
    float c1 = (float)image.At(row + 1, col)[ind]/image.At(row + 1, col)[1];
    P = (int)((c0 + c1)/2*image.At(row, col)[1]);
    }

  image.At(row, col)[ind] = image.ClampChromaPixel(P, row, col);
  }
//-----------------------------------------------------------------------------
void CornerInterpolator::InterpolatePixelInRow(Image& image, unsigned row, unsigned col, uchar ind)
  {
  assert (mp_image_info -> GetCurrentColorIndex(row, col) == 1);

  int CC, P;

  CC = 2*(int)image.At(row, col)[1] - (int)image.At(row, col - 1)[1] - (int)image.At(row, col + 1)[1];
  P = ((int)image.At(row, col - 1)[ind] + (int)image.At(row, col + 1)[ind] + CC) / 2;

  if (image.At(row, col)[1] < m_small_threshold ||
      image.At(row, col - 1)[1] < m_small_threshold ||
      image.At(row, col + 1)[1] < m_small_threshold)
    P = ((int)image.At(row, col - 1)[ind] + (int)image.At(row, col + 1)[ind] + CC) / 2;
  else
    {
    float c0 = (float)image.At(row, col - 1)[ind]/image.At(row, col - 1)[1];
    float c1 = (float)image.At(row, col + 1)[ind]/image.At(row, col + 1)[1];
    P = (int)((c0 + c1)/2*image.At(row, col)[1]);
    }

  image.At(row, col)[ind] = image.ClampChromaPixel(P, row, col);
  }
//-----------------------------------------------------------------------------
void CornerInterpolator::InterpolatePixelInNbhd(Image& image, unsigned row, unsigned col)
  {
  uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
  assert (ind != 1);

  uchar ind1 = (ind+2)%4;

  int CC[6], P[6];

  CC[0] =  2*(int)image.At(row, col)[1]
	   - (int)image.At(row - 1, col + 1)[1]
	   - (int)image.At(row + 1, col - 1)[1];
  if (image.At(row, col)[1] < m_small_threshold ||
      image.At(row - 1, col + 1)[1] < m_small_threshold ||
      image.At(row + 1, col - 1)[1] < m_small_threshold)
    P[0] = ((int)image.At(row - 1, col + 1)[ind1]
	    + (int)image.At(row + 1, col - 1)[ind1])/2
	    + CC[0] / 2;
  else
    {
    float c0 = (float)image.At(row - 1, col + 1)[ind1]/image.At(row - 1, col + 1)[1];
    float c1 = (float)image.At(row + 1, col - 1)[ind1]/image.At(row + 1, col - 1)[1];
    P[0] = (int)((c0 + c1)/2*image.At(row, col)[1]);
    }

  CC[0] = abs(CC[0])
	  + abs((int)image.At(row - 1, col + 1)[ind1]
		- (int)image.At(row + 1, col - 1)[ind1]);

  CC[1] = 2*(int)image.At(row, col)[1]
	  - (int)image.At(row - 1, col - 1)[1]
	  - (int)image.At(row + 1, col + 1)[1];
  if (image.At(row, col)[1] < m_small_threshold ||
      image.At(row - 1, col - 1)[1] < m_small_threshold ||
      image.At(row + 1, col + 1)[1] < m_small_threshold)
    P[1] = ((int)image.At(row - 1, col - 1)[ind1]
	    + (int)image.At(row + 1, col + 1)[ind1])/2
	    + CC[1] / 2;
  else
    {
    float c0 = (float)image.At(row - 1, col - 1)[ind1]/image.At(row - 1, col - 1)[1];
    float c1 = (float)image.At(row + 1, col + 1)[ind1]/image.At(row + 1, col + 1)[1];
    P[1] = (int)((c0 + c1)/2*image.At(row, col)[1]);
    }

  CC[1] = abs(CC[1])
	  + abs((int)image.At(row - 1, col - 1)[ind1]
		- (int)image.At(row + 1, col + 1)[ind1]);

  CC[2] = 2*(int)image.At(row, col)[1]
	  - (int)image.At(row, col - 1)[1]
	  - (int)image.At(row, col + 1)[1];
  if (image.At(row, col)[1] < m_small_threshold ||
      image.At(row, col - 1)[1] < m_small_threshold ||
      image.At(row, col + 1)[1] < m_small_threshold)
    P[2] = ((int)image.At(row, col - 1)[ind1]
	    + (int)image.At(row, col + 1)[ind1])/2
	    + CC[2] / 2;
  else
    {
    float c0 = (float)image.At(row, col - 1)[ind1]/image.At(row, col - 1)[1];
    float c1 = (float)image.At(row, col + 1)[ind1]/image.At(row, col + 1)[1];
    P[2] = (int)((c0 + c1)/2*image.At(row, col)[1]);
    }

  CC[2] = abs(CC[2])
	  + abs((int)image.At(row, col - 1)[ind1]
		- (int)image.At(row, col + 1)[ind1]);

  CC[3] = 2*(int)image.At(row, col)[1]
	  - (int)image.At(row - 1, col)[1]
	  - (int)image.At(row + 1, col)[1];
  if (image.At(row, col)[1] < m_small_threshold ||
      image.At(row - 1, col)[1] < m_small_threshold ||
      image.At(row + 1, col)[1] < m_small_threshold)
    P[3] = ((int)image.At(row - 1, col)[ind1]
	    + (int)image.At(row + 1, col)[ind1])/2
	    + CC[3] / 2;
  else
    {
    float c0 = (float)image.At(row - 1, col)[ind1]/image.At(row - 1, col)[1];
    float c1 = (float)image.At(row + 1, col)[ind1]/image.At(row + 1, col)[1];
    P[3] = (int)((c0 + c1)/2*image.At(row, col)[1]);
    }

  CC[3] = abs(CC[3])
	  + abs((int)image.At(row - 1, col)[ind1]
		- (int)image.At(row + 1, col)[ind1]);

  int color;

  int min_c = CC[0];
  uchar ind_c = 0;

  if (CC[0] < 300 && CC[1] < 300)
    {
    CC[4] = (2*(int)image.At(row, col)[1]
	     - (int)image.At(row - 1, col + 1)[1]
	     - (int)image.At(row + 1, col - 1)[1] +
	     + 2*(int)image.At(row, col)[1]
	     - (int)image.At(row - 1, col - 1)[1]
	     - (int)image.At(row + 1, col + 1)[1])/2;

    if (image.At(row, col)[1] < m_small_threshold ||
	image.At(row - 1, col + 1)[1] < m_small_threshold ||
	image.At(row + 1, col - 1)[1] < m_small_threshold ||
	image.At(row - 1, col - 1)[1] < m_small_threshold ||
	image.At(row + 1, col + 1)[1] < m_small_threshold)
      P[4] = ((int)image.At(row - 1, col - 1)[ind1]
	      + (int)image.At(row + 1, col + 1)[ind1]
	      + (int)image.At(row - 1, col + 1)[ind1]
	      + (int)image.At(row + 1, col - 1)[ind1])/4
	      + CC[4] / 2;
    else
      {
      float c0 = (float)image.At(row - 1, col + 1)[ind1]/image.At(row - 1, col + 1)[1];
      float c1 = (float)image.At(row + 1, col - 1)[ind1]/image.At(row + 1, col - 1)[1];
      float c2 = (float)image.At(row - 1, col - 1)[ind1]/image.At(row - 1, col - 1)[1];
      float c3 = (float)image.At(row + 1, col + 1)[ind1]/image.At(row + 1, col + 1)[1];
      P[4] = (int)((c0 + c1 + c2 + c3)/4*image.At(row, col)[1]);
      }

    min_c = CC[4];
    ind_c = 4;
    color = P[4];
    }
  else
    {
    for (uchar i = 1; i < 4; ++i)
      if (CC[i] < min_c)
	{
	min_c = CC[i];
	ind_c = i;
	}

    color = P[ind_c];
    }

  image.At(row, col)[ind1] = image.ClampChromaPixel(color, row, col);
  }
//-----------------------------------------------------------------------------
bool CornerInterpolator::Interpolate(Image& image, const RawImageInfo& image_info)
  {
  mg_log << INFO << "Running Corner interpolation" << endlog;
  mp_image_info = &image_info;

  mg_log << INFO << "Interpolating green" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
      if (ind != 1)
	InterpolateGreenPixel(image, row, col);
      }

  image.SetState(Image::FULL_GREEN_COLORS);

  mg_log << INFO << "Interpolating chroma (1st pass)" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

      if (ind == 1)
	{
	if (mp_image_info -> GetCurrentColorIndex(row, col - 1) == 0)
	  {
	  InterpolatePixelInRow(image, row, col, 0);
	  InterpolatePixelInCol(image, row, col, 2);
	  }
	else
	  {
	  InterpolatePixelInRow(image, row, col, 2);
	  InterpolatePixelInCol(image, row, col, 0);
	  }
	}
      }

  mg_log << INFO << "Interpolating chroma (2nd pass)" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

      if (ind != 1)
	InterpolatePixelInNbhd(image, row, col);
      }

  image.SetState(Image::FULL_COLORS);

  mg_log << TRACE << "Finished" << endlog;

  return true;
  }
//-----------------------------------------------------------------------------
