/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * HueGradFiltering - experimental implementation of median filtering based on
 * ``adaptive direction median'' idea. See copyrights.txt for copyright info.
******************************************************************************/


#include "stdafx.h"

#include "HueGradInterpolator.h"
#include "Image.h"
#include "RawImageInfo.h"
//-----------------------------------------------------------------------------
void HueGradInterpolator::RawFilterBy3x3AdaptiveMedian(Image& image)
  {
  mg_log << INFO << "Filtering by 3x3 adaptive median" << endlog;

  Image tmp_img(image);
  
  ulong num0 = 0, num1 = 0;

  for (unsigned row = 4; row < image.GetHeight() - 4; ++row)
    for (unsigned col = 4; col < image.GetWidth() - 4; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

      if (ind != 1)
        {
        int CC[5], P[5];

        uchar ind_c = FillGradientsForGreen(image, CC, P, row, col);

        ushort arr[9];

        if (ind_c == 2 || ind_c == 3)
          {
          ++num0;

          arr[0] = tmp_img.At(row - 2, col - 2)[ind];
          arr[1] = tmp_img.At(row, col)[ind];
          arr[2] = tmp_img.At(row + 2, col + 2)[ind];
          arr[3] = tmp_img.At(row + 2, col - 2)[ind];
          arr[4] = tmp_img.At(row - 2, col + 2)[ind];

          std::sort(&arr[0], &arr[5]);
          //At(row, col)[ind] = arr[2];
          if (image.At(row, col)[ind] == arr[4])
            image.At(row, col)[ind] = arr[3];
          else if (image.At(row, col)[ind] == arr[0])
            image.At(row, col)[ind] = arr[1];
          }
        else
          {
          ++num1;

          if (CC[0] > CC[1])
            {
            arr[0] = tmp_img.At(row - 4, col)[ind];
            arr[1] = tmp_img.At(row - 2, col)[ind];
            arr[2] = tmp_img.At(row, col)[ind];
            arr[3] = tmp_img.At(row + 2, col)[ind];
            arr[4] = tmp_img.At(row + 4, col)[ind];

            std::sort(&arr[0], &arr[5]);
            //At(row, col)[ind] = arr[2];
            if (image.At(row, col)[ind] == arr[4])
              image.At(row, col)[ind] = arr[3];
            else if (image.At(row, col)[ind] == arr[0])
              image.At(row, col)[ind] = arr[1];
            }
          else
            {
            arr[0] = tmp_img.At(row, col - 4)[ind];
            arr[1] = tmp_img.At(row, col - 2)[ind];
            arr[2] = tmp_img.At(row, col)[ind];
            arr[3] = tmp_img.At(row, col + 2)[ind];
            arr[4] = tmp_img.At(row, col + 4)[ind];

            std::sort(&arr[0], &arr[5]);
            //At(row, col)[ind] = arr[2];
            if (image.At(row, col)[ind] == arr[4])
              image.At(row, col)[ind] = arr[3];
            else if (image.At(row, col)[ind] == arr[0])
              image.At(row, col)[ind] = arr[1];
            }
          }
        }
      }

  mg_log << TRACE << "0: " << num0 << ", 1: " << num1 << endlog;
  }
//-----------------------------------------------------------------------------
void HueGradInterpolator::FullFilterBy3x3AdaptiveMedian(Image& image)
  {
  mg_log << INFO << "Full filtering by 3x3 adaptive median" << endlog;

  Image tmp_img(image);
  
  ulong num0 = 0, num1 = 0;
  int CC[30], P[30];

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      for (uchar ind = 0; ind < 3; ind += 2)
        {
        if (mp_image_info -> GetCurrentColorIndex(row, col) == ind)
          continue;

        uchar ind_c = FillGradientsForGreen(tmp_img, CC, P, row, col);

        ushort arr[9];
        
        if (ind_c < 2 || ind_c == 4)
          {
          ++num0;
          
          if (CC[0] < CC[1])
            {
            arr[0] = tmp_img.At(row, col - 2)[ind];
            arr[1] = tmp_img.At(row, col - 1)[ind];
            arr[2] = tmp_img.At(row, col + 1)[ind];
            arr[3] = tmp_img.At(row, col + 2)[ind];
            arr[4] = tmp_img.At(row, col)[ind];
            }
          else
            {
            arr[0] = tmp_img.At(row - 2, col)[ind];
            arr[1] = tmp_img.At(row - 1, col)[ind];
            arr[2] = tmp_img.At(row, col)[ind];
            arr[3] = tmp_img.At(row + 1, col)[ind];
            arr[4] = tmp_img.At(row + 2, col)[ind];
            }
          
          std::sort(&arr[0], &arr[9]);
          //At(row, col)[ind] = arr[4];
          if (image.At(row, col)[ind] >= arr[7])
            image.At(row, col)[ind] = arr[6];
          else if (image.At(row, col)[ind] <= arr[1])
            image.At(row, col)[ind] = arr[2];
          }
        else
          {
          ++num1;

          arr[0] = tmp_img.At(row - 2, col - 2)[ind];
          arr[1] = tmp_img.At(row - 1, col - 1)[ind];
          arr[2] = tmp_img.At(row, col)[ind];
          arr[3] = tmp_img.At(row + 1, col + 1)[ind];
          arr[4] = tmp_img.At(row + 2, col + 2)[ind];

          arr[5] = tmp_img.At(row + 1, col - 1)[ind];
          arr[6] = tmp_img.At(row - 1, col + 1)[ind];
          arr[7] = tmp_img.At(row + 2, col - 2)[ind];
          arr[8] = tmp_img.At(row - 2, col + 2)[ind];
          
          std::sort(&arr[0], &arr[9]);
          //At(row, col)[ind] = arr[4];
          if (image.At(row, col)[ind] >= arr[7])
            image.At(row, col)[ind] = arr[6];
          else if (image.At(row, col)[ind] <= arr[1])
            image.At(row, col)[ind] = arr[2];
          }
        }
      }

  mg_log << TRACE << "0: " << num0 << ", 1: " << num1 << endlog;
  }
//-----------------------------------------------------------------------------
