/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef HUE_GRAD_INTERPOLATOR_H
#define HUE_GRAD_INTERPOLATOR_H

#include "RawInterpolator.h"
//-----------------------------------------------------------------------------
class HueGradInterpolator: public RawInterpolator
  {
  public:
    static HueGradInterpolator& GetInterpolator();

    virtual bool Interpolate(Image& image, const RawImageInfo& image_info);

    virtual void GetParameters(Params& o_params) const;

    virtual void GetDescriptions(Params& o_params) const;

    virtual void SetParameters(Params& i_params);

  protected:
    HueGradInterpolator();

  private:
    const RawImageInfo* mp_image_info;
    int m_sigma;
    unsigned m_max_error;
    ushort m_clamp_threshold;
    ulong m_num_hue_failed, m_num_hue_ok;
    bool m_fl_smart_clamp;

  private:
    bool InterpolByHue1(int c1, int c2, int c3, int v1, int v2, int& res);

    bool InterpolByHue2(int c1, int c2, int c3, int v1, int v2, int& res);

    uchar FillGradientsForGreen(const Image& image, int CC[], int P[], unsigned row, unsigned col);

    void InterpolateGreenPixel(Image& image, unsigned row, unsigned col);

    void InterpolateChromaPixelPass1(Image& image, unsigned row, unsigned col);

    void InterpolateChromaPixelPass2(Image& image, unsigned row, unsigned col, uchar ind);

    void InterpolateChromaPixelPass3(Image& image, unsigned row, unsigned col, uchar ind);

    // Filtering stuff:
    void RawFilterBy3x3AdaptiveMedian(Image& image);

    void FullFilterBy3x3AdaptiveMedian(Image& image);
  };
//-----------------------------------------------------------------------------

#endif // HUE_GRAD_INTERPOLATOR_H
