/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * Image - main class for image storage and access.
******************************************************************************/

#include "stdafx.h"

#include "Image.h"
#include "RawImageInfo.h"
#include "RawInterpolator.h"
//-----------------------------------------------------------------------------
// ImageTmpl implementation:
//-----------------------------------------------------------------------------
template <class PixelType>
ImageTmpl<PixelType>::ImageTmpl(): m_width(0), m_height(0), m_trim_width(0),
                                   m_buffer_height(0), m_buffer_start(0), m_buffer_offs(0), m_fl_dummy_buffer(false)

  {
  }
//-----------------------------------------------------------------------------
template <class PixelType>
ImageTmpl<PixelType>::ImageTmpl(unsigned width, unsigned height, unsigned trim_width):
                                m_image(width*height),
                                m_width(width), m_height(height),
                                m_trim_width(trim_width),
                                m_buffer_height(0), m_buffer_start(0), m_buffer_offs(0), m_fl_dummy_buffer(false)
  {
  }
//-----------------------------------------------------------------------------
template <class PixelType>
void ImageTmpl<PixelType>::Create(unsigned width, unsigned height, unsigned trim_width)
  {
  m_image.resize(width*height);
  m_width = width;
  m_height = height;
  m_trim_width = trim_width;
  m_buffer_height = 0;
  m_buffer_start = 0;
  m_buffer_offs = 0;
  m_fl_dummy_buffer = false;
  }
//-----------------------------------------------------------------------------
template <class PixelType>
void ImageTmpl<PixelType>::Clear()
  {
  m_image.clear();
  m_width = 0;
  m_height = 0;
  m_trim_width = 0;
  m_buffer_height = 0;
  m_buffer_start = 0;
  m_buffer_offs = 0;
  m_buffer.clear();
  m_buffer_row_pos.clear();
  m_fl_dummy_buffer = false;
  }
//-----------------------------------------------------------------------------
template <class PixelType>
ImageTmpl<PixelType>::ImageTmpl(const ImageTmpl<PixelType>& image): m_image(image.m_width*image.m_height),
                                m_width(image.m_width), m_height(image.m_height),
                                m_trim_width(image.m_trim_width),
                                m_buffer(image.m_buffer_height*image.m_width),
                                m_buffer_row_pos(image.m_buffer_height),
                                m_buffer_height(image.m_buffer_height),
                                m_buffer_start(image.m_buffer_start),
                                m_buffer_offs(image.m_buffer_offs),
                                m_fl_dummy_buffer(image.m_fl_dummy_buffer)
  {
  m_image.assign(image.m_image.begin(), image.m_image.end());
  m_buffer.assign(image.m_buffer.begin(), image.m_buffer.end());
  m_buffer_row_pos.assign(image.m_buffer_row_pos.begin(), image.m_buffer_row_pos.end());
  }
//-----------------------------------------------------------------------------
template <class PixelType>
ImageTmpl<PixelType>& ImageTmpl<PixelType>::operator =(const ImageTmpl<PixelType>& image)
  {
  m_width = image.m_width;
  m_height = image.m_height;
  m_trim_width = image.m_trim_width;
  m_buffer_height = image.m_buffer_height;
  m_buffer_start = image.m_buffer_start;
  m_buffer_offs = image.m_buffer_offs;
  m_fl_dummy_buffer = image.m_fl_dummy_buffer;

  m_image.resize(m_width*m_height);
  m_buffer.resize(m_width*m_buffer_height);
  m_buffer_row_pos.resize(m_buffer_height);

  m_image.assign(image.m_image.begin(), image.m_image.end());
  m_buffer.assign(image.m_buffer.begin(), image.m_buffer.end());
  m_buffer_row_pos.assign(image.m_buffer_row_pos.begin(), image.m_buffer_row_pos.end());

  return *this;
  }
//-----------------------------------------------------------------------------
template <class PixelType>
void ImageTmpl<PixelType>::Set(const PixelType& val)
  {
  if (!m_width || !m_height)
    return;

  std::fill(m_image.begin(), m_image.end(), val);
  }
//-----------------------------------------------------------------------------
template <class PixelType>
void ImageTmpl<PixelType>::StartBuffer(unsigned start_row, unsigned height, bool fl_dummy_buffer)
  {
  m_buffer_height = height;
  m_buffer.resize(m_width*height);
  m_buffer_row_pos.resize(height);
  m_buffer_start = std::max(0, (int)start_row - (int)m_buffer_height + 1);
  m_buffer_offs = start_row - m_buffer_start;
  m_fl_dummy_buffer = fl_dummy_buffer;

  for (ulong i = 0; i < m_buffer_height; ++i)
    m_buffer_row_pos[i] = i*m_width;

  if (!fl_dummy_buffer)  
    for (ulong i = 0; i < m_buffer_height; ++i)
      memcpy(&m_buffer[i*m_width], &m_image[m_width*(m_buffer_start + i)], sizeof(PixelType)*m_width);
  }
//-----------------------------------------------------------------------------
template <class PixelType>
void ImageTmpl<PixelType>::FlushBufferRow()
  {
  assert(m_buffer_start < m_height);

  if (m_buffer_offs == m_buffer_height - 1)
    {
    ulong pos = m_buffer_row_pos[0];

    if (!m_fl_dummy_buffer)
      memcpy(&m_image[m_width*m_buffer_start], &m_buffer[pos], sizeof(PixelType)*m_width);

    ++m_buffer_start;

    ulong i = 0;
    for (; i < m_buffer_height - 1; ++i)
      m_buffer_row_pos[i] = m_buffer_row_pos[i + 1];

    m_buffer_row_pos[m_buffer_height - 1] = pos;

    if (!m_fl_dummy_buffer && m_buffer_start < m_height)
      memcpy(&m_buffer[pos], &m_image[m_width*(m_buffer_start + m_buffer_height - 1)], sizeof(PixelType)*m_width);
    }
  else
    ++m_buffer_offs;
  }
//-----------------------------------------------------------------------------
template <class PixelType>
void ImageTmpl<PixelType>::StopBuffer()
  {
  // Let's flush everything that is not flushed yet
  if (!m_fl_dummy_buffer)
    for (unsigned i = 0; i < m_buffer_offs; ++i)
      memcpy(&m_image[m_width*(m_buffer_start + i)], &m_buffer[m_buffer_row_pos[i]], sizeof(PixelType)*m_width);

  m_buffer_height = 0;
  m_buffer.clear();
  m_buffer_row_pos.clear();
  m_buffer_start = 0;
  m_buffer_offs = 0;
  }
//-----------------------------------------------------------------------------
// Image implementation:
//-----------------------------------------------------------------------------
Image::Image(): ImageTmpl<Pixel>(), m_clamp_val(0), m_state(UNKNOWN)
  {
  }
//-----------------------------------------------------------------------------
Image::Image(unsigned width, unsigned height, ushort clamp_val, unsigned trim_width):
             ImageTmpl<Pixel>(width, height, trim_width),
             m_clamp_val(clamp_val), m_state(UNKNOWN)
  {
  }
//-----------------------------------------------------------------------------
void Image::Create(unsigned width, unsigned height, ushort clamp_val, unsigned trim_width)
  {
  ImageTmpl<Pixel>::Create(width, height, trim_width);
  m_clamp_val = clamp_val;
  m_state = UNKNOWN;
  }
//-----------------------------------------------------------------------------
void Image::Clear()
  {
  ImageTmpl<Pixel>::Clear();
  m_clamp_val = 0;
  m_state = UNKNOWN;
  m_profile.clear();
  }
//-----------------------------------------------------------------------------
Image::Image(const Image& image): ImageTmpl<Pixel>(image),
             m_clamp_val(image.m_clamp_val), m_state(image.m_state),
             m_profile(image.m_profile)
  {
  }
//-----------------------------------------------------------------------------
Image& Image::operator =(const Image& image)
  {
  ImageTmpl<Pixel>::operator = (image);
  m_clamp_val = image.m_clamp_val;
  m_state = image.m_state;
  m_profile = image.m_profile;

  return *this;
  }
//-----------------------------------------------------------------------------
void Image::Zero()
  {
  //Pixel v;
  //v[0] = v[1] = v[2] = 0;
  //ImageTmpl<Pixel>::Set(v);
  memset(&m_image[0], 0, sizeof(Pixel)*m_image.size());
  }
//-----------------------------------------------------------------------------
template ImageTmpl<Pixel>;
template ImageTmpl<PixelCB>;
template ImageTmpl<PixelTmpl<uchar,3> >;
template ImageTmpl<PixelTmpl<float, 2> >;
template ImageTmpl<PixelTmpl<float, 3> >;
template ImageTmpl<float>;
//-----------------------------------------------------------------------------
