/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef IMAGE_TRANSFER_H
#define IMAGE_TRANSFER_H

#include "IParametrizable.h"
//-----------------------------------------------------------------------------
class Image;
//-----------------------------------------------------------------------------
class ImageTransfer: public IParametrizable
  {
  public:
    class ExImageTransferError: public MyException
      { 
      public:
	ExImageTransferError(const std::string& i_err_id, const std::string& msg): MyException("ImageTransfer", i_err_id, msg) {}
      };

  public:
    static ImageTransfer& GetTransfer();

    void Read(const std::string& file_name, Image& image);

    void Write(const std::string& file_name, const Image& image);

    void GetSupportedFormats(std::list<std::string>& formats) const;

    virtual void GetParameters(Params& o_params) const;

    virtual void GetDescriptions(Params& o_params) const;

    virtual void SetParameters(Params& i_params);

#ifdef USE_IMAGE_MAGICK
    Magick::Image* Read(const std::string& file_name);

    void Write(const std::string& file_name, Magick::Image* im_image);

    Magick::Image* ImageToIMImage(const Image& image, bool fl_trim = false);

    void IMImageToImage(Image& image, Magick::Image* im_image);
#endif

  private:
    ImageTransfer();

  private:
    int m_add_trim;
#ifdef USE_IMAGE_MAGICK
    ushort m_quality;
#endif
    std::string m_format;

  private:
    void WritePSD(const std::string& file_name, const Image& image);
  };
//-----------------------------------------------------------------------------
#endif // IMAGE_TRANSFER_H
