/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef IMAGE_UTILITIES_H
#define IMAGE_UTILITIES_H

#include "Pixel.h"
#include "WhiteBalance.h"
#include "ImageProcessor.h"
//-----------------------------------------------------------------------------
// Some methods of this class will be invoked directly, so
// they should be public
class ImageUtilities: public ImageProcessor
  {
  public:
    static ImageUtilities& GetImageProcessor();

    void Process(Image& image, const RawImageInfo& image_info, const std::string& method);

    void SubtractNoise(Image& image, ushort red, ushort green, ushort blue);

    void ApplyWB(Image& image, float red, float green, float blue);

    WhiteBalance GetWhiteBalanceFromColorTemp(float temp, float shift, const std::string& profile);

    void ApplyWBColorTemp(Image& image, float temp, float shift, const std::string& profile);

    void EmbedColorProfile(Image& image, const std::string& name);

    void ApplyColorProfile(Image& image, const std::string& in_profile, const std::string& out_profile);

    void CorrectExposure(Image& image, float coef);

    void ConvertRGB2RGGB(Image& image);

    void ConvertRGGB2RGB(Image& image);

    void ApplyGamma(Image& image, float gamma);

    void BuildHistogram(std::vector<PixelLong>& histogram, ulong size, const Image& image, bool fl_count_zeroes);

    virtual void GetParameters(Params& o_params) const;

    virtual void GetDescriptions(Params& o_params) const;

    virtual void SetParameters(Params& i_params);

  private:
    ImageUtilities();

  private:
    float m_gamma, m_expo_correction;
  };
//-----------------------------------------------------------------------------

#endif // IMAGE_UTILITIES_H

