/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * MedianInterpolator - implementation of Freeman median interpolation,
 * based on description given in ``Interpolation Methods for the Bayer
 * Color Array'' by Rajeev Ramanath. See copyrights.txt for patent info.
******************************************************************************/

#include "stdafx.h"

#include "MedianInterpolator.h"
#include "BilinearInterpolator.h"
#include "Image.h"
#include "RawImageInfo.h"
#include "logstream.h"
//-----------------------------------------------------------------------------
MedianInterpolator::MedianInterpolator(): IParametrizable("MedianInterpolator")
  {
  m_dim = 1;
  }
//-----------------------------------------------------------------------------
MedianInterpolator& MedianInterpolator::GetInterpolator()
  {
  static MedianInterpolator m_int;
  return m_int;
  }
//-----------------------------------------------------------------------------
void MedianInterpolator::GetParameters(Params& o_params) const
  {
  o_params["MedianInt_NbhdSize"] = boost::lexical_cast<std::string>(m_dim);
  }
//-----------------------------------------------------------------------------
void MedianInterpolator::GetDescriptions(Params& o_params) const
  {
  o_params["MedianInt_NbhdSize"] = "Size of neighborhood considered during the interpolation";
  }
//-----------------------------------------------------------------------------
void MedianInterpolator::SetParameters(Params& i_params)
  {
  if (i_params.exist("MedianInt_NbhdSize"))
    m_dim = boost::lexical_cast<ulong>(i_params.pop("MedianInt_NbhdSize"));
  }
//-----------------------------------------------------------------------------
void MedianInterpolator::InterpolatePixel(const Image& image, Image& res_img, unsigned row, unsigned col,
                                          uchar ind, uchar ind_b)
  {
  std::vector<int> diffs;
  diffs.reserve((m_dim*2 + 1) * (m_dim*2 + 1));

  for (unsigned r = row - m_dim; r <= row + m_dim; ++r)
    for (unsigned c = col - m_dim; c <= col + m_dim; ++c)
      diffs.push_back((int)image.At(r, c)[ind] - (int)image.At(r, c)[ind_b]);

  std::sort(diffs.begin(), diffs.end());

  int color = diffs[(m_dim*2 + 1) * (m_dim*2 + 1) / 2] + (int)image.At(row, col)[ind_b];

  if (color < 0)
    color = 0;

  if (color > image.GetClampVal())
    color = image.GetClampVal();

  res_img.At(row, col)[ind] = color;
  }
//-----------------------------------------------------------------------------
bool MedianInterpolator::Interpolate(Image& image, const RawImageInfo& image_info)
  {
  // First we need to obtain preliminary interpolation ...
  Image tmp_img(image);

  BilinearInterpolator::GetInterpolator().Interpolate(tmp_img, image_info);

  mg_log << INFO << "Running Median interpolation" << endlog;
  mp_image_info = &image_info;

  for (unsigned row = m_dim; row < image.GetHeight() - m_dim; ++row)
    for (unsigned col = m_dim; col < image.GetWidth() - m_dim; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

      InterpolatePixel(tmp_img, image, row, col, (ind+1)%3, ind);
      InterpolatePixel(tmp_img, image, row, col, (ind+2)%3, ind);
      }

  image.SetState(Image::FULL_COLORS);

  mg_log << TRACE << "Finished" << endlog;

  return true;
  }
//-----------------------------------------------------------------------------