/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * MultiGradInterpolator - implementation of multi-gradient interpolation,
 * based on description given in ``Color Filter Array Recovery Using a
 * Threshold-based Variable Number of Gradients'' by Ed Chang, Shiufun Cheung
 * and Davis Pan. See copyrights.txt for patent info.
******************************************************************************/

#include "stdafx.h"

#include "MultiGradientInterpolator.h"
#include "Image.h"
#include "RawImageInfo.h"
#include "logstream.h"
//-----------------------------------------------------------------------------
MultiGradientInterpolator::MultiGradientInterpolator(): IParametrizable("MultiGradientInterpolator")
  {
  }
//-----------------------------------------------------------------------------
MultiGradientInterpolator& MultiGradientInterpolator::GetInterpolator()
  {
  static MultiGradientInterpolator mg_int;
  return mg_int;
  }
//-----------------------------------------------------------------------------
void MultiGradientInterpolator::GetParameters(Params& o_params) const
  {
  }
//-----------------------------------------------------------------------------
void MultiGradientInterpolator::GetDescriptions(Params& o_params) const
  {
  }
//-----------------------------------------------------------------------------
void MultiGradientInterpolator::SetParameters(Params& i_params)
  {
  }
//-----------------------------------------------------------------------------
void MultiGradientInterpolator::GetGreenGradients(Image& image, unsigned row, unsigned col,
						  int grds[8],
						  uchar i0, uchar i2)
  {
  grds[0] = abs((int)image.At(row - 1, col)[i0] - (int)image.At(row + 1, col)[i0]) + abs((int)image.At(row - 2, col)[1] - (int)image.At(row, col)[1]) + abs((int)image.At(row - 1, col - 1)[1] - (int)image.At(row + 1, col  - 1)[1]) / 2 + abs((int)image.At(row - 1, col + 1)[1] - (int)image.At(row + 1, col + 1)[1]) / 2 + abs((int)image.At(row - 2, col - 1)[i2] - (int)image.At(row, col - 1)[i2]) / 2 + abs((int)image.At(row - 2, col + 1)[i2] - (int)image.At(row, col + 1)[i2]) / 2;
  grds[1] = abs((int)image.At(row, col + 1)[i2] - (int)image.At(row, col - 1)[i2]) + abs((int)image.At(row, col + 2)[1] - (int)image.At(row, col)[1]) + abs((int)image.At(row - 1, col + 1)[1] - (int)image.At(row - 1, col - 1)[1]) / 2 + abs((int)image.At(row + 1, col + 1)[1] - (int)image.At(row + 1, col  - 1)[1]) / 2 + abs((int)image.At(row - 1, col + 2)[i0] - (int)image.At(row - 1, col)[i0]) / 2 + abs((int)image.At(row + 1, col + 2)[i0] - (int)image.At(row + 1, col)[i0]) / 2;
  grds[2] = abs((int)image.At(row + 1, col)[i0] - (int)image.At(row - 1, col)[i0]) + abs((int)image.At(row + 2, col)[1] - (int)image.At(row, col)[1]) + abs((int)image.At(row + 1, col  - 1)[1] - (int)image.At(row - 1, col - 1)[1]) / 2 + abs((int)image.At(row + 1, col + 1)[1] - (int)image.At(row - 1, col + 1)[1]) / 2 + abs((int)image.At(row + 2, col - 1)[i2] - (int)image.At(row, col - 1)[i2]) / 2 + abs((int)image.At(row + 2, col + 1)[i2] - (int)image.At(row, col + 1)[i2]) / 2;
  grds[3] = abs((int)image.At(row, col - 1)[i2] - (int)image.At(row, col + 1)[i2]) + abs((int)image.At(row, col - 2)[1] - (int)image.At(row, col)[1]) + abs((int)image.At(row - 1, col - 1)[1] - (int)image.At(row - 1, col + 1)[1]) / 2 + abs((int)image.At(row + 1, col  - 1)[1] - (int)image.At(row + 1, col + 1)[1]) / 2 + abs((int)image.At(row - 1, col - 2)[i0] - (int)image.At(row - 1, col)[i0]) / 2 + abs((int)image.At(row + 1, col - 2)[i0] - (int)image.At(row + 1, col)[i0]) / 2;
  grds[4] = abs((int)image.At(row - 1, col + 1)[1] - (int)image.At(row + 1, col  - 1)[1]) + abs((int)image.At(row - 2, col + 2)[1] - (int)image.At(row, col)[1]) + abs((int)image.At(row - 2, col + 1)[i2] - (int)image.At(row, col - 1)[i2]) + abs((int)image.At(row - 1, col + 2)[i0] - (int)image.At(row + 1, col)[i0]);
  grds[5] = abs((int)image.At(row + 1, col + 1)[1] - (int)image.At(row - 1, col - 1)[1]) + abs((int)image.At(row + 2, col + 2)[1] - (int)image.At(row, col)[1]) + abs((int)image.At(row + 2, col + 1)[i2] - (int)image.At(row, col - 1)[i2]) + abs((int)image.At(row + 1, col + 2)[i0] - (int)image.At(row - 1, col)[i0]);
  grds[6] = abs((int)image.At(row - 1, col - 1)[1] - (int)image.At(row + 1, col + 1)[1]) + abs((int)image.At(row - 2, col - 2)[1] - (int)image.At(row, col)[1]) + abs((int)image.At(row - 2, col - 1)[i2] - (int)image.At(row, col + 1)[i2]) + abs((int)image.At(row - 1, col - 2)[i0] - (int)image.At(row + 1, col)[i0]);
  grds[7] = abs((int)image.At(row + 1, col  - 1)[1] - (int)image.At(row - 1, col + 1)[1]) + abs((int)image.At(row + 2, col - 2)[1] - (int)image.At(row, col)[1]) + abs((int)image.At(row + 2, col - 1)[i2] - (int)image.At(row, col + 1)[i2]) + abs((int)image.At(row + 1, col - 2)[i0] - (int)image.At(row - 1, col)[i0]);
  }
//-----------------------------------------------------------------------------
void MultiGradientInterpolator::GetNonGreenGradients(Image& image, unsigned row, unsigned col,
						     int grds[8],
						     uchar i0, uchar i2)
  {
  grds[0] = abs((int)image.At(row - 1, col)[1] - (int)image.At(row + 1, col)[1]) + abs((int)image.At(row - 2, col)[i0] - (int)image.At(row, col)[i0]) + abs((int)image.At(row - 1, col - 1)[i2] - (int)image.At(row + 1, col - 1)[i2]) / 2 + abs((int)image.At(row - 1, col + 1)[i2] - (int)image.At(row + 1, col + 1)[i2]) / 2 + abs((int)image.At(row - 2, col - 1)[1] - (int)image.At(row, col - 1)[1]) / 2 + abs((int)image.At(row - 2, col + 1)[1] - (int)image.At(row, col + 1)[1]) / 2;
  grds[1] = abs((int)image.At(row, col + 1)[1] - (int)image.At(row, col - 1)[1]) + abs((int)image.At(row, col + 2)[i0] - (int)image.At(row, col)[i0]) + abs((int)image.At(row - 1, col + 1)[i2] - (int)image.At(row - 1, col - 1)[i2]) / 2 + abs((int)image.At(row + 1, col + 1)[i2] - (int)image.At(row + 1, col - 1)[i2]) / 2 + abs((int)image.At(row - 1, col + 2)[1] - (int)image.At(row - 1, col)[1]) / 2 + abs((int)image.At(row + 1, col + 2)[1] - (int)image.At(row + 1, col)[1]) / 2;
  grds[2] = abs((int)image.At(row + 1, col)[1] - (int)image.At(row - 1, col)[1]) + abs((int)image.At(row + 2, col)[i0] - (int)image.At(row, col)[i0]) + abs((int)image.At(row + 1, col - 1)[i2] - (int)image.At(row - 1, col - 1)[i2]) / 2 + abs((int)image.At(row + 1, col + 1)[i2] - (int)image.At(row - 1, col + 1)[i2]) / 2 + abs((int)image.At(row + 2, col - 1)[1] - (int)image.At(row, col - 1)[1]) / 2 + abs((int)image.At(row + 2, col + 1)[1] - (int)image.At(row, col + 1)[1]) / 2;
  grds[3] = abs((int)image.At(row, col - 1)[1] - (int)image.At(row, col + 1)[1]) + abs((int)image.At(row, col - 2)[i0] - (int)image.At(row, col)[i0]) + abs((int)image.At(row - 1, col - 1)[i2] - (int)image.At(row - 1, col + 1)[i2]) / 2 + abs((int)image.At(row + 1, col - 1)[i2] - (int)image.At(row + 1, col + 1)[i2]) / 2 + abs((int)image.At(row - 1, col - 2)[1] - (int)image.At(row - 1, col)[1]) / 2 + abs((int)image.At(row + 1, col - 2)[1] - (int)image.At(row + 1, col)[1]) / 2;
  grds[4] = abs((int)image.At(row - 1, col + 1)[i2] - (int)image.At(row + 1, col - 1)[i2]) + abs((int)image.At(row - 2, col + 2)[i0] - (int)image.At(row, col)[i0]) + abs((int)image.At(row - 1, col)[1] - (int)image.At(row, col - 1)[1]) / 2 + abs((int)image.At(row, col + 1)[1] - (int)image.At(row + 1, col)[1]) / 2 + abs((int)image.At(row - 2, col + 1)[1] - (int)image.At(row - 1, col)[1]) / 2 + abs((int)image.At(row - 1, col + 2)[1] - (int)image.At(row, col + 1)[1]) / 2;
  grds[5] = abs((int)image.At(row + 1, col + 1)[i2] - (int)image.At(row - 1, col - 1)[i2]) + abs((int)image.At(row + 2, col + 2)[i0] - (int)image.At(row, col)[i0]) + abs((int)image.At(row, col + 1)[1] - (int)image.At(row - 1, col)[1]) / 2 + abs((int)image.At(row + 1, col)[1] - (int)image.At(row, col - 1)[1]) / 2 + abs((int)image.At(row + 1, col + 2)[1] - (int)image.At(row, col + 1)[1]) / 2 + abs((int)image.At(row + 2, col + 1)[1] - (int)image.At(row + 1, col)[1]) / 2;
  grds[6] = abs((int)image.At(row - 1, col - 1)[i2] - (int)image.At(row + 1, col + 1)[i2]) + abs((int)image.At(row - 2, col - 2)[i0] - (int)image.At(row, col)[i0]) + abs((int)image.At(row - 1, col)[1] - (int)image.At(row, col + 1)[1]) / 2 + abs((int)image.At(row, col - 1)[1] - (int)image.At(row + 1, col)[1]) / 2 + abs((int)image.At(row - 2, col - 1)[1] - (int)image.At(row - 1, col)[1]) / 2 + abs((int)image.At(row - 1, col - 2)[1] - (int)image.At(row, col - 1)[1]) / 2;
  grds[7] = abs((int)image.At(row + 1, col - 1)[i2] - (int)image.At(row - 1, col + 1)[i2]) + abs((int)image.At(row + 2, col - 2)[i0] - (int)image.At(row, col)[i0]) + abs((int)image.At(row, col - 1)[1] - (int)image.At(row - 1, col)[1]) / 2 + abs((int)image.At(row + 1, col)[1] - (int)image.At(row, col + 1)[1]) / 2 + abs((int)image.At(row + 1, col - 2)[1] - (int)image.At(row, col - 1)[1]) / 2 + abs((int)image.At(row + 2, col - 1)[1] - (int)image.At(row + 1, col)[1]) / 2;  
  }
//-----------------------------------------------------------------------------
void MultiGradientInterpolator::GetGreenSum(Image& image, unsigned row, unsigned col, uchar ind,
					    uchar i0, uchar i2, int pixel[3])
  {
  switch (ind)
    {
    case 0:
      pixel[0] = (int)image.At(row - 1, col)[i0];
      pixel[1] = ((int)image.At(row - 2, col)[1] + (int)image.At(row, col)[1]) / 2;
      pixel[2] = ((int)image.At(row - 2, col - 1)[i2] + (int)image.At(row - 2, col + 1)[i2] + (int)image.At(row, col - 1)[i2] + (int)image.At(row, col + 1)[i2]) / 4;
      break;
      
    case 1:
      pixel[0] = ((int)image.At(row - 1, col)[i0] + (int)image.At(row - 1, col + 2)[i0] + (int)image.At(row + 1, col)[i0] + (int)image.At(row + 1, col + 2)[i0]) / 4;
      pixel[1] = ((int)image.At(row, col + 2)[1] + (int)image.At(row, col)[1]) / 2;
      pixel[2] = (int)image.At(row, col + 1)[i2];
      break;
      
    case 2:
      pixel[0] = (int)image.At(row + 1, col)[i0];
      pixel[1] = ((int)image.At(row, col)[1] + (int)image.At(row + 2, col)[1]) / 2;
      pixel[2] = ((int)image.At(row, col - 1)[i2] + (int)image.At(row, col + 1)[i2] + (int)image.At(row + 2, col - 1)[i2] + (int)image.At(row + 2, col + 1)[i2]) / 4;
      break;
      
    case 3:
      pixel[0] = ((int)image.At(row - 1, col - 2)[i0] + (int)image.At(row - 1, col)[i0] + (int)image.At(row + 1, col - 2)[i0] + (int)image.At(row + 1, col)[i0]) / 4;
      pixel[1] = ((int)image.At(row, col - 2)[1] + (int)image.At(row, col)[1]) / 2;
      pixel[2] = (int)image.At(row, col - 1)[i2];
      break;
      
    case 4:
      pixel[0] = ((int)image.At(row - 1, col)[i0] + (int)image.At(row - 1, col + 2)[i0]) / 2;
      pixel[1] = (int)image.At(row - 1, col + 1)[1];
      pixel[2] = ((int)image.At(row - 2, col + 1)[i2] + (int)image.At(row, col + 1)[i2]) / 2;
      break;
      
    case 5:
      pixel[0] = ((int)image.At(row + 1, col)[i0] + (int)image.At(row + 1, col + 2)[i0]) / 2;
      pixel[1] = (int)image.At(row + 1, col + 1)[1];
      pixel[2] = ((int)image.At(row, col + 1)[i2] + (int)image.At(row + 2, col + 1)[i2]) / 2;
      break;
      
    case 6:
      pixel[0] = ((int)image.At(row - 1, col - 2)[i0] + (int)image.At(row - 1, col)[i0]) / 2;
      pixel[1] = (int)image.At(row - 1, col - 1)[1];
      pixel[2] = ((int)image.At(row - 2, col - 1)[i2] + (int)image.At(row, col - 1)[i2]) / 2;
      break;
      
    case 7:
      pixel[0] = ((int)image.At(row + 1, col - 2)[i0] + (int)image.At(row + 1, col)[i0]) / 2;
      pixel[1] = (int)image.At(row + 1, col  - 1)[1];
      pixel[2] = ((int)image.At(row, col - 1)[i2] + (int)image.At(row + 2, col - 1)[i2]) / 2;
      break;
    }
  }
//-----------------------------------------------------------------------------
void MultiGradientInterpolator::GetNonGreenSum(Image& image, unsigned row, unsigned col, uchar ind,
					       uchar i0, uchar i2, int pixel[3])
  {
  switch (ind)
    {
    case 0:
      pixel[0] = ((int)image.At(row - 2, col)[i0] + (int)image.At(row, col)[i0]) / 2;
      pixel[1] = (int)image.At(row - 1, col)[1];
      pixel[2] = ((int)image.At(row - 1, col - 1)[i2] + (int)image.At(row - 1, col + 1)[i2]) / 2;
      break;
      
    case 1:
      pixel[0] = ((int)image.At(row, col)[i0] + (int)image.At(row, col + 2)[i0]) / 2;
      pixel[1] = (int)image.At(row, col + 1)[1];
      pixel[2] = ((int)image.At(row - 1, col + 1)[i2] + (int)image.At(row + 1, col + 1)[i2]) / 2;
      break;
      
    case 2:
      pixel[0] = ((int)image.At(row, col)[i0] + (int)image.At(row + 2, col)[i0]) / 2;
      pixel[1] = (int)image.At(row + 1, col)[1];
      pixel[2] = ((int)image.At(row + 1, col - 1)[i2] + (int)image.At(row + 1, col + 1)[i2]) / 2;
      break;
      
    case 3:
      pixel[0] = ((int)image.At(row, col - 2)[i0] + (int)image.At(row, col)[i0]) / 2;
      pixel[1] = (int)image.At(row, col - 1)[1];
      pixel[2] = ((int)image.At(row - 1, col - 1)[i2] + (int)image.At(row + 1, col - 1)[i2]) / 2;
      break;
      
    case 4:
      pixel[0] = ((int)image.At(row - 2, col + 2)[i0] + (int)image.At(row, col)[i0]) / 2;
      pixel[1] = ((int)image.At(row - 2, col + 1)[1] + (int)image.At(row - 1, col)[1] + (int)image.At(row - 1, col + 2)[1] + (int)image.At(row, col + 1)[1]) / 4;
      pixel[2] = (int)image.At(row - 1, col + 1)[i2];
      break;
      
    case 5:
      pixel[0] = ((int)image.At(row, col)[i0] + (int)image.At(row + 2, col + 2)[i0]) / 2;
      pixel[1] = ((int)image.At(row, col + 1)[1] + (int)image.At(row + 1, col)[1] + (int)image.At(row + 1, col + 2)[1] + (int)image.At(row + 2, col + 1)[1]) / 4;
      pixel[2] = (int)image.At(row + 1, col + 1)[i2];
      break;
      
    case 6:
      pixel[0] = ((int)image.At(row - 2, col - 2)[i0] + (int)image.At(row, col)[i0]) / 2;
      pixel[1] = ((int)image.At(row - 2, col - 1)[1] + (int)image.At(row - 1, col - 2)[1] + (int)image.At(row - 1, col)[1] + (int)image.At(row, col - 1)[1]) / 4;
      pixel[2] = (int)image.At(row - 1, col - 1)[i2];
      break;
      
    case 7:
      pixel[0] = ((int)image.At(row, col)[i0] + (int)image.At(row + 2, col - 2)[i0]) / 2;
      pixel[1] = ((int)image.At(row, col - 1)[1] + (int)image.At(row + 1, col - 2)[1] + (int)image.At(row + 1, col)[1] + (int)image.At(row + 2, col - 1)[1]) / 4;
      pixel[2] = (int)image.At(row + 1, col - 1)[i2];
      break;
    }
  }
//-----------------------------------------------------------------------------
void MultiGradientInterpolator::InterpolatePixel(Image& image, unsigned row, unsigned col)
  {
  uchar cur_ind = mp_image_info -> GetCurrentColorIndex(row, col),
	above_ind = mp_image_info -> GetCurrentColorIndex(row - 1, col);

  int grds[8];

  switch (cur_ind)
    {
    case 0: // Red
      GetNonGreenGradients(image, row, col, grds, 0, 2);
      break;

    case 1: // Green
      if (above_ind == 0)
	      GetGreenGradients(image, row, col, grds, 0, 2);
      else
	      GetGreenGradients(image, row, col, grds, 2, 0);
      break;

    case 2: // Blue
      GetNonGreenGradients(image, row, col, grds, 2, 0);
      break;
    }

  // Let's find min/max
  int min_grad = *std::min_element(&grds[0], &grds[8]);
  int max_grad = *std::max_element(&grds[0], &grds[8]);

  float k1 = 1.5, k2 = 0.5;
  int threshold = (int)(k1 * min_grad  + k2 * (max_grad - min_grad));

  std::vector<uchar> ok_grads;

  for (uchar i = 0; i < 8; ++i)
    if (grds[i] <= threshold)
      ok_grads.push_back(i);

  int pixel[3], sum_pixel[3] = {0, 0, 0};

  for (std::vector<uchar>::const_iterator it = ok_grads.begin();
       it != ok_grads.end(); ++it)
    {
    switch (cur_ind)
      {
      case 0: // Red
	    GetNonGreenSum(image, row, col, *it, 0, 2, pixel);
	    break;

      case 1: // Green
	    if (above_ind == 0)
	      GetGreenSum(image, row, col, *it, 0, 2, pixel);
	    else
	      {
	      GetGreenSum(image, row, col, *it, 2, 0, pixel);
	      std::swap(pixel[0], pixel[2]);
	      }
	    break;

      case 2: // Blue
	    GetNonGreenSum(image, row, col, *it, 2, 0, pixel);
	    std::swap(pixel[0], pixel[2]);
	    break;
      }

    sum_pixel[0] += pixel[0];
    sum_pixel[1] += pixel[1];
    sum_pixel[2] += pixel[2];
    }

  sum_pixel[0] /= (int)ok_grads.size();
  sum_pixel[1] /= (int)ok_grads.size();
  sum_pixel[2] /= (int)ok_grads.size();

  int new_pixel[3];

  switch (cur_ind)
    {
    case 0: // Red
      new_pixel[0] = image.At(row, col)[0];
      new_pixel[1] = (int)image.At(row, col)[0] + (sum_pixel[1] - sum_pixel[0]);
      new_pixel[2] = (int)image.At(row, col)[0] + (sum_pixel[2] - sum_pixel[0]);
      break;

    case 1: // Green
      new_pixel[0] = (int)image.At(row, col)[1] + (sum_pixel[0] - sum_pixel[1]);
      new_pixel[1] = image.At(row, col)[1];
      new_pixel[2] = (int)image.At(row, col)[1] + (sum_pixel[2] - sum_pixel[1]);
      break;

    case 2: // Blue
      new_pixel[0] = (int)image.At(row, col)[2] + (sum_pixel[0] - sum_pixel[2]);
      new_pixel[1] = (int)image.At(row, col)[2] + (sum_pixel[1] - sum_pixel[2]);
      new_pixel[2] = image.At(row, col)[2];
      break;
    }

  image.At(row, col)[0] = image.ClampChromaPixel(new_pixel[0], row, col);
  image.At(row, col)[1] = image.ClampGreenPixel(new_pixel[1], row, col);
  image.At(row, col)[2] = image.ClampChromaPixel(new_pixel[2], row, col);
  }
//-----------------------------------------------------------------------------
bool MultiGradientInterpolator::Interpolate(Image& image, const RawImageInfo& image_info)
  {
  mg_log << INFO << "Running Multi-Gradient interpolation" << endlog;
  mp_image_info = &image_info;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      InterpolatePixel(image, row, col);

  image.SetState(Image::FULL_COLORS);

  mg_log << TRACE << "Finished" << endlog;

  return true;
  }
//-----------------------------------------------------------------------------
