/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"

#include "ParamsDispatcher.h"
#include "IParametrizable.h"
//-----------------------------------------------------------------------------
ParamsDispatcher::ParamsDispatcher()
  {
  }
//-----------------------------------------------------------------------------
ParamsDispatcher& ParamsDispatcher::GetDispatcher()
  {
  static ParamsDispatcher disp;
  return disp;
  }
//-----------------------------------------------------------------------------
void ParamsDispatcher::Register(const std::string& module, IParametrizable* ip_client)
  {
  m_clients.push_back(ClientInfo(module, ip_client));
  }
//-----------------------------------------------------------------------------
void ParamsDispatcher::UnRegister(IParametrizable* ip_client)
  {
  for (ClientsList::iterator it = m_clients.begin();
       it != m_clients.end(); ++it)
    if (it -> second == ip_client)
      {
      m_clients.erase(it);
      return;
      }
  }
//-----------------------------------------------------------------------------

void ParamsDispatcher::GetParameters(ParamsList& o_params) const
  {
  for (ClientsList::const_iterator it = m_clients.begin();
       it != m_clients.end(); ++it)
    {
    Params params;
    it -> second -> GetParameters(params);
    o_params.push_back(ModuleInfo(it -> first, params));
    }
  }
//-----------------------------------------------------------------------------
void ParamsDispatcher::GetDescriptions(ParamsList& o_params) const
  {
  for (ClientsList::const_iterator it = m_clients.begin();
       it != m_clients.end(); ++it)
    {
    Params params;
    it -> second -> GetDescriptions(params);
    o_params.push_back(ModuleInfo(it -> first, params));
    }

  }
//-----------------------------------------------------------------------------
void ParamsDispatcher::SetParameters(Params& i_params) const
  {
  try
    {
    for (ClientsList::const_iterator it = m_clients.begin();
         it != m_clients.end(); ++it)
      it -> second -> SetParameters(i_params);
    }
  catch (const boost::bad_lexical_cast& ex)
    {
    throw Params::ExBadParameter("WrongValueFormat", ex.what());
    }

  if (!i_params.empty())
    {
    mg_log << ERR << "Some parameters were rejected (wrong names?)" << endlog;
    mg_log << ERR << "Here is the list of them:" << endlog;

    for (Params::iterator it = i_params.begin(); it != i_params.end(); ++it)
      mg_log << ERR << it -> first << " = " << it -> second << endlog;
    }
  }
//-----------------------------------------------------------------------------
