/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef PIXEL_H
#define PIXEL_H

//-----------------------------------------------------------------------------
// We rely on the 1-byte alignment in the code
#pragma pack(1)
//-----------------------------------------------------------------------------
template <class T = ushort, uchar col_num = 3>
struct PixelTmpl
  {
  T m_colors[col_num];

  T operator [](uchar ind) const { return m_colors[ind]; }
  T& operator[](uchar ind) { return m_colors[ind]; }

  bool operator == (const PixelTmpl<T>& pix) const
    {
    for (uchar i = 0; i < col_num; ++i)
      if (m_colors[i] == pix.m_colors[i])
	return false;

    return true;
    }

  bool operator != (const PixelTmpl<T>& pix) const
    {
    return !((*this) == pix);
    }

  T Length() const
    {
    T sum;
    for (uchar i = 0; i < col_num; ++i)
      sum += m_colors[i]*m_colors[i];

    return sqrt(sum);
    }

  void Zero()
    {
    for (uchar i = 0; i < col_num; ++i)
      m_colors[i] = 0;
    }

// friend PixelTmpl<T> operator +(const PixelTmpl<T>& pix0, const PixelTmpl<T>& pix1);
//  friend PixelTmpl<T> operator -(const PixelTmpl<T>& pix0, const PixelTmpl<T>& pix1);
  };
//-----------------------------------------------------------------------------
/*
template <class T, uchar col_num>
inline PixelTmpl<T, col_num> operator +(const PixelTmpl<T, col_num>& pix0, const PixelTmpl<T, col_num>& pix1)
  {
  PixelTmpl<T, col_num> res;
  for (uchar i = 0; i < col_num; ++i)
    res[i] = pix0[i] + pix1[i];

  return res;
  }
//-----------------------------------------------------------------------------
template <class T, uchar col_num>
inline PixelTmpl<T, col_num> operator -(const PixelTmpl<T, col_num>& pix0, const PixelTmpl<T, col_num>& pix1)
  {
  PixelTmpl<T, col_num> res;
  for (uchar i = 0; i < col_num; ++i)
    res[i] = pix0[i] - pix1[i];

  return res;
  }*/
//-----------------------------------------------------------------------------
typedef PixelTmpl<ushort> Pixel;
typedef PixelTmpl<float> PixelFloat;
typedef PixelTmpl<ulong> PixelLong;
//-----------------------------------------------------------------------------
class PixelCB: public PixelTmpl<double, 4>
  {
  public:
    PixelCB() {}

    PixelCB(const PixelCB& pix)
      {
      m_colors[0] = pix.m_colors[0];
      m_colors[1] = pix.m_colors[1];
      m_colors[2] = pix.m_colors[2];
      m_colors[3] = pix.m_colors[3];
      }

    PixelCB(const Pixel& pix)
      {
      m_colors[3] = sqrt((double)pix[0]*(double)pix[0] +
			 (double)pix[1]*(double)pix[1] +
			 (double)pix[2]*(double)pix[2]);
      if (m_colors[3] > 0)
	{
	m_colors[0] = pix[0]/m_colors[3];
	m_colors[1] = pix[1]/m_colors[3];
	m_colors[2] = pix[2]/m_colors[3];
	}
      else
	{
	m_colors[0] = 1/3.0;
	m_colors[1] = 1/3.0;
	m_colors[2] = 1/3.0;
	}
      };

    operator Pixel() const
      {
      Pixel pix;
      int v = (int)(m_colors[0] * m_colors[3]);

      if (v < 0)
	v = 0;
      else if (v > 65535)
	v = 65535;
      pix[0] = v;
      
      v = (int)(m_colors[1] * m_colors[3]);
      if (v < 0)
	v = 0;
      else if (v > 65535)
	v = 65535;
      pix[1] = v;

      v = (int)(m_colors[2] * m_colors[3]);
      if (v < 0)
	v = 0;
      else if (v > 65535)
	v = 65535;
      pix[2] = v;

      return pix;
      };

    void Normalize(unsigned clamp_val = 65535)
      {
      if (m_colors[3] < 0)
	m_colors[3] = 0;
      else if (m_colors[3] > clamp_val*sqrt(3.0))
	m_colors[3] = clamp_val*sqrt(3.0);

      if (m_colors[0] < 0)
	m_colors[0] = 0;

      if (m_colors[1] < 0)
	m_colors[1] = 0;

      if (m_colors[2] < 0)
	m_colors[2] = 0;

      double sum = sqrt(m_colors[0]*m_colors[0] + m_colors[1]*m_colors[1] + m_colors[2]*m_colors[2]);
      if (sum > 0)
	{
	m_colors[0] /= sum;
	m_colors[1] /= sum;
	m_colors[2] /= sum;
	}
      else
	{
	m_colors[0] = 1/3.0;
	m_colors[1] = 1/3.0;
	m_colors[2] = 1/3.0;
	}
      };

  double ChDist2(const PixelCB& pix) const
    {
    return sqr(m_colors[0] - pix[0])
	   + sqr(m_colors[1] - pix[1])
	   + sqr(m_colors[2] - pix[2]);
    }
  };
//-----------------------------------------------------------------------------
// This is not real XYZ color system!!!
// (just a part of it - to help in WB transformations)
class Pixel_xyz: public PixelFloat
  {
  public:
    Pixel_xyz() {}

    Pixel_xyz(const Pixel_xyz& pix)
      {
      m_colors[0] = pix.m_colors[0];
      m_colors[1] = pix.m_colors[1];
      m_colors[2] = pix.m_colors[2];
      }

    PixelFloat GetRGB() const
      {
      PixelFloat pix;

      pix[0] = 2.3646138465383655*m_colors[0]
	       - 0.896540570739668*m_colors[1]
	       - 0.4680732757986974*m_colors[2];

      pix[1] = -0.515166208447888*m_colors[0]
	       + 1.4264081038563887*m_colors[1]
	       + 0.08875810459149917*m_colors[2];

      pix[2] = 0.005203699075231191*m_colors[0]
	       - 0.014408162665216048*m_colors[1]
	       + 1.0092044635899848*m_colors[2];

      return pix;
      };
  };
//-----------------------------------------------------------------------------
#pragma pack()
//-----------------------------------------------------------------------------
#endif // PIXEL_H

