/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef PROCESSOR_H
#define PROCESSOR_H

#include "Image.h"
#include "IParametrizable.h"
//-----------------------------------------------------------------------------
class RawInterpolator;
class RawFileReader;
class RawImageInfo;
class WBCalculator;
class ImageProcessor;
//-----------------------------------------------------------------------------
class Processor: public IParametrizable
  {
  public:
    class ExProcessorError: public MyException
      { 
      public:
        ExProcessorError(const std::string& i_err_id, const std::string& msg): MyException("Processor", i_err_id, msg) {}
      };

  public:
    ~Processor();

    static Processor& GetProcessor();

    void SetFileToLoad(const std::string& file_name)
      {
      m_file_to_load = file_name;
      }

    void LoadImage(const std::string& file_name, bool fl_parse_only = false,
                   bool fl_small_icon_mode = false);

    void Process(const std::string& cmds);

    const Image& GetImage() const;

    Image& GetImage();

    const std::vector<Image>& GetThumbnails() const;

    const RawImageInfo& GetImageInfo() const;

    virtual void GetParameters(Params& o_params) const;

    virtual void GetDescriptions(Params& o_params) const;

    virtual void SetParameters(Params& i_params);

    static void DumpSupportedFileFormats();

    static void DumpDefaultParameters();

    static void DumpParametersDescriptions();

  private:
    Image m_image;
    std::vector<Image> m_thumbnails;
    std::map<std::string, RawInterpolator*> m_interpolators;
    std::map<std::string, RawFileReader*> m_readers;
    std::map<std::string, WBCalculator*> m_wb_calculators;
    std::map<std::string, ImageProcessor*> m_image_processors;
    RawImageInfo* mp_image_info;
    std::string m_file_to_load;

  private:
    Processor();
  };
//-----------------------------------------------------------------------------
inline const Image& Processor::GetImage() const
  {
  return m_image;
  }
//-----------------------------------------------------------------------------
inline Image& Processor::GetImage()
  {
  return m_image;
  }
//-----------------------------------------------------------------------------
inline const std::vector<Image>& Processor::GetThumbnails() const
  {
  return m_thumbnails;
  }
//-----------------------------------------------------------------------------
inline const RawImageInfo& Processor::GetImageInfo() const
  {
  return *mp_image_info;
  }
//-----------------------------------------------------------------------------
#endif // PROCESSOR_H