/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef RAW_FILE_READER_H
#define RAW_FILE_READER_H

#include "IParametrizable.h"
//-----------------------------------------------------------------------------
class RawImageInfo;
class Image;
//-----------------------------------------------------------------------------
class RawFileReader: public virtual IParametrizable
  {
  public:
    class ExRawFileError: public MyException
      { 
      public:
        ExRawFileError(const std::string& i_err_id, const std::string& msg): MyException("RawFileReader", i_err_id, msg) {}
      };

    virtual void LoadRaw(const std::string& file_name, Image& image, bool fl_parse_only = false,
                         bool fl_small_icon_mode = false);

    virtual void GetParameters(Params& o_params) const;

    virtual void GetDescriptions(Params& o_params) const;

    virtual void SetParameters(Params& i_params);

    virtual RawImageInfo& GetImageInfo() = 0;

    virtual RawImageInfo* RetrieveImageInfo() = 0;

    virtual void CleanUp();

  protected:
    RawFileReader();

    virtual void ParseCiffTag(ushort type, ulong len, ulong abs_offset) {}

    virtual void Load(Image& p_image) = 0;

    ushort Get2();
    
    ulong Get4();

    virtual void PrepareImageInfo() = 0;

    void Seek(std::ifstream::off_type offs, std::ifstream::seekdir way);

  protected:
    std::ifstream m_file;
    ushort m_order;
    ulong m_trim_width, m_file_length;
    bool m_fl_leave_border, m_fl_use_standard_size,
         m_fl_parse_only, m_fl_small_icon_mode;

  private:
    void ParseCiff(ulong offset, ulong length);
  };
//-----------------------------------------------------------------------------
inline void RawFileReader::Seek(std::ifstream::off_type offs, std::ifstream::seekdir way)
  {
  if (offs < 0 || offs > m_file_length)
    {
    CleanUp();
    throw ExRawFileError("IncorrectRAWFormat", "This RAW file has incorrect or unsupported format");
    }

  m_file.seekg(offs, way);
  }
//-----------------------------------------------------------------------------
#endif // RAW_FILE_READER_H
