/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef RAW_IMAGE_INFO_H
#define RAW_IMAGE_INFO_H

//-----------------------------------------------------------------------------
class RawFileReader;
//-----------------------------------------------------------------------------
class RawImageInfo
  {
  friend class RawFileReader;
  friend class Processor;

  public:
    typedef std::vector<char> Thumbnail;
    typedef std::vector<Thumbnail> ThumbnailsList;

  public:
    RawImageInfo();

    virtual ~RawImageInfo() {}

    unsigned GetWidth() const;

    unsigned GetHeight() const;

    double GetRedWB() const;

    double GetGreenWB() const;

    double GetBlueWB() const;

    const std::string& GetModelName() const;

    const std::string& GetManufacturerName() const;

    const std::string& GetOwnerName() const;

    double GetNoise(uchar num) const;

    ushort GetMaxColor(uchar num) const;

    uchar GetCurrentColorIndex(unsigned row, unsigned col) const;

    unsigned GetLeftBorder() const;

    unsigned GetTopBorder() const;

    const ThumbnailsList& GetThumbnails() const;

    int GetRotationAngle() const;

  protected:
    std::string m_manufacturer_name, m_model_name, m_owner_name;
    unsigned m_width, m_height, m_std_width, m_std_height;
    float m_aspect_ratio;
    int m_rotation_angle;
    double m_red_wb, m_blue_wb;
    double m_noise[3];
    ushort m_max_color[3];
    unsigned m_left_border, m_top_border;
    uchar m_filters;
    ThumbnailsList m_thumbnails;
  };
//-----------------------------------------------------------------------------
inline RawImageInfo::RawImageInfo()
  {
  m_red_wb = m_blue_wb = 0.0;
  }
//-----------------------------------------------------------------------------
inline double RawImageInfo::GetRedWB() const
  {
  return m_red_wb;
  }
//-----------------------------------------------------------------------------
inline double RawImageInfo::GetGreenWB() const
  {
  return 1.0;
  }
//-----------------------------------------------------------------------------
inline double RawImageInfo::GetBlueWB() const
  {
  return m_blue_wb;
  }
//-----------------------------------------------------------------------------
inline const std::string& RawImageInfo::GetModelName() const
  {
  return m_model_name;
  }
//-----------------------------------------------------------------------------
inline const std::string& RawImageInfo::GetManufacturerName() const
  {
  return m_manufacturer_name;
  }
//-----------------------------------------------------------------------------
inline const std::string& RawImageInfo::GetOwnerName() const
  {
  return m_owner_name;
  }
//-----------------------------------------------------------------------------
inline unsigned RawImageInfo::GetWidth() const
  {
  return m_width;
  }
//-----------------------------------------------------------------------------
inline unsigned RawImageInfo::GetHeight() const
  {
  return m_height;
  }
//-----------------------------------------------------------------------------
inline double RawImageInfo::GetNoise(uchar num) const
  {
  return m_noise[num];
  }
//-----------------------------------------------------------------------------
inline ushort RawImageInfo::GetMaxColor(uchar num) const
  {
  return m_max_color[num];
  }
//-----------------------------------------------------------------------------
inline uchar RawImageInfo::GetCurrentColorIndex(unsigned row, unsigned col) const
  {
  //return (m_filters >> ((((row) << 1 & 14) + ((col) & 1)) << 1) & 3);
  //return (row & 1) + (col & 1);
  return ((m_filters >> ((row&1)<<2)) >> ((col&1)<<1)) & 3;
  }
//-----------------------------------------------------------------------------
inline unsigned RawImageInfo::GetLeftBorder() const
  {
  return m_left_border;
  }
//-----------------------------------------------------------------------------
inline unsigned RawImageInfo::GetTopBorder() const
  {
  return m_top_border;
  }
//-----------------------------------------------------------------------------
inline int RawImageInfo::GetRotationAngle() const
  {
  return m_rotation_angle;
  }
//-----------------------------------------------------------------------------

#endif // RAW_IMAGE_INFO_H
