/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * TVImageProcessor - implementation of noise filtering using
 * Total Variation approach performed separately for chroma components and
 * for brightness, based on description given in ``The digital TV filter and
 * nonlinear denoising'' by Tony F. Chan, Stanley Osher and Jianhong Shen.
******************************************************************************/

#include "stdafx.h"

#include "TVImageProcessor.h"
#include "Image.h"
//-----------------------------------------------------------------------------
TVImageProcessor::TVImageProcessor(): IParametrizable("TVImageProcessor")
  {
  /* ISO 400
  m_q_br = 28000;
  m_q_ch[0] = m_q_ch[1] = m_q_ch[2] = 0.01;

  m_lambda_br = 1/sqr(m_q_br);
  m_lambda_ch[0] = 1/sqr(m_q_ch[0]);
  m_lambda_ch[1] = 1/sqr(m_q_ch[1]);
  m_lambda_ch[2] = 1/sqr(m_q_ch[2]);
  m_alpha = 1e-6;
  m_num_of_iter = 50;
  */
  // ISO 50
  /*
  m_q_br = 2000;
  m_q_ch[0] = m_q_ch[1] = m_q_ch[2] = 0.005;
  */

  // Estimated values:
  // ISO 50:
  m_q_ch[0] = 0.004;
  m_q_ch[1] = 0.004;
  m_q_ch[2] = 0.004;
  m_q_br = 1245;
  m_lambda_br = 0.02;
  m_lambda_ch = 4;


  // ISO 100:
  /*
  m_q_ch[0] = 0.005;
  m_q_ch[1] = 0.005;
  m_q_ch[2] = 0.005;
  m_q_br = 1455;
  */

  // ISO 200:
  /*
  m_q_ch[0] = 0.007;
  m_q_ch[1] = 0.007;
  m_q_ch[2] = 0.007;
  m_q_br = 1870;
  */

  // ISO 400:
  /*
  m_q_ch[0] = 0.01;
  m_q_ch[1] = 0.01;
  m_q_ch[2] = 0.01;
  m_q_br = 2865;
  m_lambda_br = 0.002;
  m_lambda_ch = 0.1;
  */

  //m_lambda_br = 0.005;
  /*
  m_lambda_br = 1/sqr(m_q_br);
  m_lambda_ch[0] = 1/sqr(m_q_ch[0]);
  m_lambda_ch[1] = 1/sqr(m_q_ch[1]);
  m_lambda_ch[2] = 1/sqr(m_q_ch[2]);
  */
  m_alpha = 1e-6;
  m_num_of_iter = 4;
  }
//-----------------------------------------------------------------------------
TVImageProcessor& TVImageProcessor::GetImageProcessor()
  {
  static TVImageProcessor tv_ip;
  return tv_ip;
  }
//-----------------------------------------------------------------------------
void TVImageProcessor::GetParameters(Params& o_params) const
  {
  o_params["TV_Iterations"] = boost::lexical_cast<std::string>(m_num_of_iter);
  o_params["TV_AmountChroma"] = boost::lexical_cast<std::string>(m_lambda_ch);
  o_params["TV_AmountBrightness"] = boost::lexical_cast<std::string>(m_lambda_br);
  }
//-----------------------------------------------------------------------------
void TVImageProcessor::GetDescriptions(Params& o_params) const
  {
  o_params["TVImageProcessor::ApplyTV()"] = "Total Variation image filter (denoising)";

  o_params["TV_AmountChroma"] = "Strength of chroma preservation (lower values give more denoising)";
  o_params["TV_AmountBrightness"] = "Strength of brightness preservation (lower values give more denoising)";
  o_params["TV_Iterations"] = "Number of TV iterations";
  }
//-----------------------------------------------------------------------------
void TVImageProcessor::SetParameters(Params& i_params)
  {
  if (i_params.exist("TV_AmountChroma"))
    m_lambda_ch = boost::lexical_cast<float>(i_params.pop("TV_AmountChroma"));
  if (i_params.exist("TV_AmountBrightness"))
    m_lambda_br = boost::lexical_cast<float>(i_params.pop("TV_AmountBrightness"));
  if (i_params.exist("TV_Iterations"))
    m_num_of_iter = boost::lexical_cast<ulong>(i_params.pop("TV_Iterations"));
  }
//-----------------------------------------------------------------------------
void TVImageProcessor::Process(Image& image, const RawImageInfo& image_info, const std::string& method)
  {
  if (method == "ApplyTV()")
    ApplyTV(image, image_info);
  else throw ExImageProcessorError("UnknMethod", "Attempting to call unexisting processing method TVImageProcessor::" + method);
  }
//-----------------------------------------------------------------------------
inline
void TVImageProcessor::GetGradNorm(PixelGrad& pix,
                                   const ImageCB& img_cb,
                                   unsigned row, unsigned col)
  {
  PixelCB orig(img_cb.AtBuffer(row, col));

  float sum = orig.ChDist2(img_cb.AtBuffer(row, col - 1))
              + orig.ChDist2(img_cb.AtBuffer(row - 1, col))
              + orig.ChDist2(img_cb.AtBuffer(row, col + 1))
              + orig.ChDist2(img_cb.AtBuffer(row + 1, col));

  pix[0] = 1/sqrt(sum + m_alpha);

  sum = sqr(orig[3] - img_cb.AtBuffer(row, col - 1)[3])
        + sqr(orig[3] - img_cb.AtBuffer(row - 1, col)[3])
        + sqr(orig[3] - img_cb.AtBuffer(row, col + 1)[3])
        + sqr(orig[3] - img_cb.AtBuffer(row + 1, col)[3]);

  pix[1] = 1/sqrt(sum + m_alpha);
  }
//-----------------------------------------------------------------------------
inline
void TVImageProcessor::ApplyStep(const Image& orig_img,
                                 Image& image,
                                 const ImageCB& img_cb,
                                 const GradMatrix& grad,
                                 unsigned row, unsigned col)
  {
  //   0 
  // 1 2 3
  //   4
  PixelCB pix[5] = {img_cb.AtBuffer(row - 1, col),
                    img_cb.AtBuffer(row, col - 1),
                    img_cb.AtBuffer(row, col),
                    img_cb.AtBuffer(row, col + 1),
                    img_cb.AtBuffer(row + 1, col)};

  PixelCB orig_pix(orig_img.At(row, col));

  float grads_ch[5] = {grad.AtBuffer(row - 1, col)[0],
                       grad.AtBuffer(row, col - 1)[0],
                       grad.AtBuffer(row, col)[0],
                       grad.AtBuffer(row, col + 1)[0],
                       grad.AtBuffer(row + 1, col)[0]};

  float grads_br[5] = {grad.AtBuffer(row - 1, col)[1],
                       grad.AtBuffer(row, col - 1)[1],
                       grad.AtBuffer(row, col)[1],
                       grad.AtBuffer(row, col + 1)[1],
                       grad.AtBuffer(row + 1, col)[1]};

  float sum_ch = m_lambda_ch + 4*grads_ch[2] + grads_ch[0] + grads_ch[1]
                 + grads_ch[3] + grads_ch[4];

  float h_ch[5] = {(grads_ch[2] + grads_ch[0]) / sum_ch,
                  (grads_ch[2] + grads_ch[1]) / sum_ch,
                  m_lambda_ch / sum_ch,
                  (grads_ch[2] + grads_ch[3]) / sum_ch,
                  (grads_ch[2] + grads_ch[4]) / sum_ch};

  float sum_br = m_lambda_br + 4*grads_br[2] + grads_br[0] + grads_br[1]
                 + grads_br[3] + grads_br[4];

  float h_br[5] = {(grads_br[2] + grads_br[0]) / sum_br,
                   (grads_br[2] + grads_br[1]) / sum_br,
                   m_lambda_br / sum_br,
                   (grads_br[2] + grads_br[3]) / sum_br,
                   (grads_br[2] + grads_br[4]) / sum_br};

  for (uchar i = 0; i < 3; ++i)
    pix[2][i] = h_ch[0] * pix[0][i]
                + h_ch[1] * pix[1][i]
                + h_ch[2] * orig_pix[i]
                + h_ch[3] * pix[3][i]
                + h_ch[4] * pix[4][i];

  pix[2][3] = h_br[0] * pix[0][3]
              + h_br[1] * pix[1][3]
              + h_br[2] * orig_pix[3]
              + h_br[3] * pix[3][3]
              + h_br[4] * pix[4][3];

  pix[2].Normalize();
  image.AtBuffer(row, col) = pix[2];
  }
//-----------------------------------------------------------------------------
void TVImageProcessor::EstimateNoise(const Image& image)
  {
  PixelCB sum_pix;

  sum_pix.Zero();

  for (unsigned row = 0; row < image.GetHeight(); ++row)
    for (unsigned col = 0; col < image.GetWidth(); ++col)
      {
      PixelCB pix = image.At(row, col);
      sum_pix[0] += pix[0];
      sum_pix[1] += pix[1];
      sum_pix[2] += pix[2];
      sum_pix[3] += pix[3];
      }

  ulong size = image.GetHeight()*image.GetWidth();
  sum_pix[0] /= size;
  sum_pix[1] /= size;
  sum_pix[2] /= size;
  sum_pix[3] /= size;

  mg_log << INFO << "Estimation: avg=" << sum_pix[0] << ", "
    << sum_pix[1] << ", "
    << sum_pix[2] << ", "
    << sum_pix[3] << endlog;

  PixelCB mean, dev, dev2;

  dev.Zero(); dev2.Zero();

  for (unsigned row = 0; row < image.GetHeight(); ++row)
    for (unsigned col = 0; col < image.GetWidth(); ++col)
      {
      PixelCB pix = image.At(row, col);
      pix[0] -= sum_pix[0];
      pix[1] -= sum_pix[1];
      pix[2] -= sum_pix[2];
      pix[3] -= sum_pix[3];

      mean[0] += pix[0];
      mean[1] += pix[1];
      mean[2] += pix[2];
      mean[3] += pix[3];

      dev[0] += pix[0]; dev2[0] += pix[0]*pix[0];
      dev[1] += pix[1]; dev2[1] += pix[1]*pix[1];
      dev[2] += pix[2]; dev2[2] += pix[2]*pix[2];
      dev[3] += pix[3]; dev2[3] += pix[3]*pix[3];
      }

  mean[0] /= size;
  mean[1] /= size;
  mean[2] /= size;
  mean[3] /= size;

  dev2[0] /= size;
  dev2[1] /= size;
  dev2[2] /= size;
  dev2[3] /= size;

  dev2[0] = sqrt(dev2[0]);
  dev2[1] = sqrt(dev2[1]);
  dev2[2] = sqrt(dev2[2]);
  dev2[3] = sqrt(dev2[3]);

  mg_log << INFO << "Estimation: mean=" << mean[0] << ", "
    << mean[1] << ", "
    << mean[2] << ", "
    << mean[3] << endlog;

  mg_log << INFO << "Estimation: dev2=" << dev2[0] << ", "
    << dev2[1] << ", "
    << dev2[2] << ", "
    << dev2[3] << endlog;
  }
//-----------------------------------------------------------------------------
void TVImageProcessor::ApplyTV(Image& image, const RawImageInfo& image_info)
  {
  mp_image_info = &image_info;

  ImageCB img_cb(image.GetWidth(), 0);
  Image orig_img(image);
  GradMatrix grad(image.GetWidth(), 0);

  for (ulong num = 0; num < m_num_of_iter; ++num)
    {
    mg_log << INFO << "Applying TV (pass " << num+1 << " of " << m_num_of_iter << ")" << endlog;

    mg_log << DEBUG << "Lambdas: " << m_lambda_ch << ", "
      << m_lambda_br << endlog;

    img_cb.StartBuffer(0, 5, true);
    for (unsigned row = 0; row < 4; ++row)
      {
      for (unsigned col = 0; col < image.GetWidth(); ++col)
        img_cb.AtBuffer(row, col) = image.At(row, col);

      img_cb.FlushBufferRow();
      }

    grad.StartBuffer(1, 3, true);
    for (unsigned row = 1; row < 3; ++row)
      {
      for (unsigned col = 1; col < image.GetWidth() - 1; ++col)
        GetGradNorm(grad.AtBuffer(row, col), img_cb, row, col);

      grad.FlushBufferRow();
      }

    m_est_lambda[0] = m_est_lambda[1]
      = m_est_lambda[2] = m_est_lambda[3] = 0;

    image.StartBuffer(2, 3);
    for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
      {
      for (unsigned col = 0; col < image.GetWidth(); ++col)
        img_cb.AtBuffer(row + 2, col) = image.At(row + 2, col);

      for (unsigned col = 1; col < image.GetWidth() - 1; ++col)
        GetGradNorm(grad.AtBuffer(row + 1, col), img_cb, row + 1, col);

      for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
        ApplyStep(orig_img, image, img_cb, grad, row, col);

      image.FlushBufferRow();
      img_cb.FlushBufferRow();
      grad.FlushBufferRow();
      }

    image.StopBuffer();
    img_cb.StopBuffer();
    grad.StopBuffer();

/*
    if (num > 0)
      {
      ulong size = (image.GetWidth() - 4)*(image.GetHeight() - 4);

      m_lambda_ch[0] = 1/sqr(m_q_ch[0]) * m_est_lambda[0] / size;
      m_lambda_ch[1] = 1/sqr(m_q_ch[1]) * m_est_lambda[1] / size;
      m_lambda_ch[2] = 1/sqr(m_q_ch[2]) * m_est_lambda[2] / size;
      m_lambda_br = 1/sqr(m_q_br) * m_est_lambda[3] / size;
      }
*/
    }
  }
//-----------------------------------------------------------------------------