/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef TV_IMAGE_PROCESSOR_H
#define TV_IMAGE_PROCESSOR_H

#include "ImageProcessor.h"
#include "Image.h"
//-----------------------------------------------------------------------------
class TVImageProcessor: public ImageProcessor
  {
  public:
    static TVImageProcessor& GetImageProcessor();

    void Process(Image& image, const RawImageInfo& image_info, const std::string& method);

    void ApplyTV(Image& image, const RawImageInfo& image_info);

    virtual void GetParameters(Params& o_params) const;

    virtual void GetDescriptions(Params& o_params) const;

    virtual void SetParameters(Params& i_params);

  //protected:
    TVImageProcessor();

  //private:
    const RawImageInfo* mp_image_info;
    float m_lambda_br, m_lambda_ch, m_lambda, m_est_lambda[4], m_alpha,
          m_q_br, m_q_ch[3];
    ulong m_num_of_iter;
    typedef PixelTmpl<float, 2> PixelGrad;
    typedef ImageTmpl<PixelGrad> GradMatrix;

  private:
    void GetGradNorm(PixelGrad& pix, const ImageCB& img_cb, unsigned row, unsigned col);

    void ApplyStep(const Image& orig_img, Image& image,
                   const ImageCB& img_cb,
                   const GradMatrix& grad, unsigned row, unsigned col);

    void EstimateNoise(const Image& image);
  };
//-----------------------------------------------------------------------------
#endif // TV_IMAGE_PROCESSOR_H
