/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * WBCalculator - base class for white balance calculation.
******************************************************************************/

#include "stdafx.h"

#include <deque>

#include <lcms.h>

#include "WBCalculator.h"
#include "RawImageInfo.h"
#include "ImageTransfer.h"

#define WRITE_SEGMENTED_IMAGE "segmented.tif"
//#undef WRITE_SEGMENTED_IMAGE
//-----------------------------------------------------------------------------
PlanckianCalculator::PlanckianPoint PlanckianCalculator::m_planckian_points[] = {
PlanckianCalculator::PlanckianPoint(0.6499, 0.3474, 1000),
PlanckianCalculator::PlanckianPoint(0.6361, 0.3594, 1100),
PlanckianCalculator::PlanckianPoint(0.6226, 0.3703, 1200),
PlanckianCalculator::PlanckianPoint(0.6095, 0.3801, 1300),
PlanckianCalculator::PlanckianPoint(0.5966, 0.3887, 1400),
PlanckianCalculator::PlanckianPoint(0.5841, 0.3962, 1500),
PlanckianCalculator::PlanckianPoint(0.5720, 0.4025, 1600),
PlanckianCalculator::PlanckianPoint(0.5601, 0.4076, 1700),
PlanckianCalculator::PlanckianPoint(0.5486, 0.4118, 1800),
PlanckianCalculator::PlanckianPoint(0.5375, 0.4150, 1900),
PlanckianCalculator::PlanckianPoint(0.5267, 0.4173, 2000),
PlanckianCalculator::PlanckianPoint(0.5162, 0.4188, 2100),
PlanckianCalculator::PlanckianPoint(0.5062, 0.4196, 2200),
PlanckianCalculator::PlanckianPoint(0.4965, 0.4198, 2300),
PlanckianCalculator::PlanckianPoint(0.4872, 0.4194, 2400),
PlanckianCalculator::PlanckianPoint(0.4782, 0.4186, 2500),
PlanckianCalculator::PlanckianPoint(0.4696, 0.4173, 2600),
PlanckianCalculator::PlanckianPoint(0.4614, 0.4158, 2700),
PlanckianCalculator::PlanckianPoint(0.4535, 0.4139, 2800),
PlanckianCalculator::PlanckianPoint(0.4460, 0.4118, 2900),
PlanckianCalculator::PlanckianPoint(0.4388, 0.4095, 3000),
PlanckianCalculator::PlanckianPoint(0.4320, 0.4070, 3100),
PlanckianCalculator::PlanckianPoint(0.4254, 0.4044, 3200),
PlanckianCalculator::PlanckianPoint(0.4192, 0.4018, 3300),
PlanckianCalculator::PlanckianPoint(0.4132, 0.3990, 3400),
PlanckianCalculator::PlanckianPoint(0.4075, 0.3962, 3500),
PlanckianCalculator::PlanckianPoint(0.4021, 0.3934, 3600),
PlanckianCalculator::PlanckianPoint(0.3969, 0.3905, 3700),
PlanckianCalculator::PlanckianPoint(0.3919, 0.3877, 3800),
PlanckianCalculator::PlanckianPoint(0.3872, 0.3849, 3900),
PlanckianCalculator::PlanckianPoint(0.3827, 0.3820, 4000),
PlanckianCalculator::PlanckianPoint(0.3784, 0.3793, 4100),
PlanckianCalculator::PlanckianPoint(0.3743, 0.3765, 4200),
PlanckianCalculator::PlanckianPoint(0.3704, 0.3738, 4300),
PlanckianCalculator::PlanckianPoint(0.3666, 0.3711, 4400),
PlanckianCalculator::PlanckianPoint(0.3631, 0.3685, 4500),
PlanckianCalculator::PlanckianPoint(0.3596, 0.3659, 4600),
PlanckianCalculator::PlanckianPoint(0.3563, 0.3634, 4700),
PlanckianCalculator::PlanckianPoint(0.3532, 0.3609, 4800),
PlanckianCalculator::PlanckianPoint(0.3502, 0.3585, 4900),
PlanckianCalculator::PlanckianPoint(0.3473, 0.3561, 5000),
PlanckianCalculator::PlanckianPoint(0.3446, 0.3538, 5100),
PlanckianCalculator::PlanckianPoint(0.3419, 0.3516, 5200),
PlanckianCalculator::PlanckianPoint(0.3394, 0.3494, 5300),
PlanckianCalculator::PlanckianPoint(0.3369, 0.3472, 5400),
PlanckianCalculator::PlanckianPoint(0.3346, 0.3451, 5500),
PlanckianCalculator::PlanckianPoint(0.3323, 0.3431, 5600),
PlanckianCalculator::PlanckianPoint(0.3302, 0.3411, 5700),
PlanckianCalculator::PlanckianPoint(0.3281, 0.3392, 5800),
PlanckianCalculator::PlanckianPoint(0.3261, 0.3373, 5900),
PlanckianCalculator::PlanckianPoint(0.3242, 0.3355, 6000),
PlanckianCalculator::PlanckianPoint(0.3223, 0.3337, 6100),
PlanckianCalculator::PlanckianPoint(0.3205, 0.3319, 6200),
PlanckianCalculator::PlanckianPoint(0.3188, 0.3302, 6300),
PlanckianCalculator::PlanckianPoint(0.3171, 0.3286, 6400),
PlanckianCalculator::PlanckianPoint(0.3155, 0.3270, 6500),
PlanckianCalculator::PlanckianPoint(0.3140, 0.3254, 6600),
PlanckianCalculator::PlanckianPoint(0.3125, 0.3238, 6700),
PlanckianCalculator::PlanckianPoint(0.3110, 0.3224, 6800),
PlanckianCalculator::PlanckianPoint(0.3097, 0.3209, 6900),
PlanckianCalculator::PlanckianPoint(0.3083, 0.3195, 7000),
PlanckianCalculator::PlanckianPoint(0.3070, 0.3181, 7100),
PlanckianCalculator::PlanckianPoint(0.3058, 0.3168, 7200),
PlanckianCalculator::PlanckianPoint(0.3045, 0.3154, 7300),
PlanckianCalculator::PlanckianPoint(0.3034, 0.3142, 7400),
PlanckianCalculator::PlanckianPoint(0.3022, 0.3129, 7500),
PlanckianCalculator::PlanckianPoint(0.3011, 0.3117, 7600),
PlanckianCalculator::PlanckianPoint(0.3000, 0.3105, 7700),
PlanckianCalculator::PlanckianPoint(0.2990, 0.3094, 7800),
PlanckianCalculator::PlanckianPoint(0.2980, 0.3082, 7900),
PlanckianCalculator::PlanckianPoint(0.2970, 0.3071, 8000),
PlanckianCalculator::PlanckianPoint(0.2961, 0.3061, 8100),
PlanckianCalculator::PlanckianPoint(0.2952, 0.3050, 8200),
PlanckianCalculator::PlanckianPoint(0.2943, 0.3040, 8300),
PlanckianCalculator::PlanckianPoint(0.2934, 0.3030, 8400),
PlanckianCalculator::PlanckianPoint(0.2926, 0.3020, 8500),
PlanckianCalculator::PlanckianPoint(0.2917, 0.3011, 8600),
PlanckianCalculator::PlanckianPoint(0.2910, 0.3001, 8700),
PlanckianCalculator::PlanckianPoint(0.2902, 0.2992, 8800),
PlanckianCalculator::PlanckianPoint(0.2894, 0.2983, 8900),
PlanckianCalculator::PlanckianPoint(0.2887, 0.2975, 9000),
PlanckianCalculator::PlanckianPoint(0.2880, 0.2966, 9100),
PlanckianCalculator::PlanckianPoint(0.2873, 0.2958, 9200),
PlanckianCalculator::PlanckianPoint(0.2866, 0.2950, 9300),
PlanckianCalculator::PlanckianPoint(0.2860, 0.2942, 9400),
PlanckianCalculator::PlanckianPoint(0.2853, 0.2934, 9500),
PlanckianCalculator::PlanckianPoint(0.2847, 0.2927, 9600),
PlanckianCalculator::PlanckianPoint(0.2841, 0.2919, 9700),
PlanckianCalculator::PlanckianPoint(0.2835, 0.2912, 9800),
PlanckianCalculator::PlanckianPoint(0.2829, 0.2905, 9900),
PlanckianCalculator::PlanckianPoint(0.2824, 0.2898, 10000),
PlanckianCalculator::PlanckianPoint(0.2818, 0.2891, 10100),
PlanckianCalculator::PlanckianPoint(0.2813, 0.2884, 10200),
PlanckianCalculator::PlanckianPoint(0.2807, 0.2878, 10300),
PlanckianCalculator::PlanckianPoint(0.2802, 0.2871, 10400),
PlanckianCalculator::PlanckianPoint(0.2797, 0.2865, 10500),
PlanckianCalculator::PlanckianPoint(0.2792, 0.2859, 10600),
PlanckianCalculator::PlanckianPoint(0.2788, 0.2853, 10700),
PlanckianCalculator::PlanckianPoint(0.2783, 0.2847, 10800),
PlanckianCalculator::PlanckianPoint(0.2778, 0.2841, 10900),
PlanckianCalculator::PlanckianPoint(0.2774, 0.2836, 11000),
PlanckianCalculator::PlanckianPoint(0.2770, 0.2830, 11100),
PlanckianCalculator::PlanckianPoint(0.2765, 0.2825, 11200),
PlanckianCalculator::PlanckianPoint(0.2761, 0.2819, 11300),
PlanckianCalculator::PlanckianPoint(0.2757, 0.2814, 11400),
PlanckianCalculator::PlanckianPoint(0.2753, 0.2809, 11500),
PlanckianCalculator::PlanckianPoint(0.2749, 0.2804, 11600),
PlanckianCalculator::PlanckianPoint(0.2745, 0.2799, 11700),
PlanckianCalculator::PlanckianPoint(0.2742, 0.2794, 11800),
PlanckianCalculator::PlanckianPoint(0.2738, 0.2789, 11900),
PlanckianCalculator::PlanckianPoint(0.2734, 0.2785, 12000),
PlanckianCalculator::PlanckianPoint(0.2731, 0.2780, 12100),
PlanckianCalculator::PlanckianPoint(0.2727, 0.2776, 12200),
PlanckianCalculator::PlanckianPoint(0.2724, 0.2771, 12300),
PlanckianCalculator::PlanckianPoint(0.2721, 0.2767, 12400),
PlanckianCalculator::PlanckianPoint(0.2717, 0.2763, 12500),
PlanckianCalculator::PlanckianPoint(0.2714, 0.2758, 12600),
PlanckianCalculator::PlanckianPoint(0.2711, 0.2754, 12700),
PlanckianCalculator::PlanckianPoint(0.2708, 0.2750, 12800),
PlanckianCalculator::PlanckianPoint(0.2705, 0.2746, 12900),
PlanckianCalculator::PlanckianPoint(0.2702, 0.2742, 13000),
PlanckianCalculator::PlanckianPoint(0.2699, 0.2738, 13100),
PlanckianCalculator::PlanckianPoint(0.2696, 0.2735, 13200),
PlanckianCalculator::PlanckianPoint(0.2694, 0.2731, 13300),
PlanckianCalculator::PlanckianPoint(0.2691, 0.2727, 13400),
PlanckianCalculator::PlanckianPoint(0.2688, 0.2724, 13500),
PlanckianCalculator::PlanckianPoint(0.2686, 0.2720, 13600),
PlanckianCalculator::PlanckianPoint(0.2683, 0.2717, 13700),
PlanckianCalculator::PlanckianPoint(0.2680, 0.2713, 13800),
PlanckianCalculator::PlanckianPoint(0.2678, 0.2710, 13900),
PlanckianCalculator::PlanckianPoint(0.2675, 0.2707, 14000),
PlanckianCalculator::PlanckianPoint(0.2673, 0.2703, 14100),
PlanckianCalculator::PlanckianPoint(0.2671, 0.2700, 14200),
PlanckianCalculator::PlanckianPoint(0.2668, 0.2697, 14300),
PlanckianCalculator::PlanckianPoint(0.2666, 0.2694, 14400),
PlanckianCalculator::PlanckianPoint(0.2664, 0.2691, 14500),
PlanckianCalculator::PlanckianPoint(0.2662, 0.2688, 14600),
PlanckianCalculator::PlanckianPoint(0.2659, 0.2685, 14700),
PlanckianCalculator::PlanckianPoint(0.2657, 0.2682, 14800),
PlanckianCalculator::PlanckianPoint(0.2655, 0.2679, 14900),
PlanckianCalculator::PlanckianPoint(0.2653, 0.2676, 15000),
PlanckianCalculator::PlanckianPoint(0.2651, 0.2673, 15100),
PlanckianCalculator::PlanckianPoint(0.2649, 0.2671, 15200),
PlanckianCalculator::PlanckianPoint(0.2647, 0.2668, 15300),
PlanckianCalculator::PlanckianPoint(0.2645, 0.2665, 15400),
PlanckianCalculator::PlanckianPoint(0.2643, 0.2663, 15500),
PlanckianCalculator::PlanckianPoint(0.2641, 0.2660, 15600),
PlanckianCalculator::PlanckianPoint(0.2639, 0.2657, 15700),
PlanckianCalculator::PlanckianPoint(0.2638, 0.2655, 15800),
PlanckianCalculator::PlanckianPoint(0.2636, 0.2652, 15900),
PlanckianCalculator::PlanckianPoint(0.2634, 0.2650, 16000),
PlanckianCalculator::PlanckianPoint(0.2632, 0.2648, 16100),
PlanckianCalculator::PlanckianPoint(0.2631, 0.2645, 16200),
PlanckianCalculator::PlanckianPoint(0.2629, 0.2643, 16300),
PlanckianCalculator::PlanckianPoint(0.2627, 0.2641, 16400),
PlanckianCalculator::PlanckianPoint(0.2626, 0.2638, 16500),
PlanckianCalculator::PlanckianPoint(0.2624, 0.2636, 16600),
PlanckianCalculator::PlanckianPoint(0.2622, 0.2634, 16700),
PlanckianCalculator::PlanckianPoint(0.2621, 0.2632, 16800),
PlanckianCalculator::PlanckianPoint(0.2619, 0.2629, 16900),
PlanckianCalculator::PlanckianPoint(0.2618, 0.2627, 17000),
PlanckianCalculator::PlanckianPoint(0.2616, 0.2625, 17100),
PlanckianCalculator::PlanckianPoint(0.2615, 0.2623, 17200),
PlanckianCalculator::PlanckianPoint(0.2613, 0.2621, 17300),
PlanckianCalculator::PlanckianPoint(0.2612, 0.2619, 17400),
PlanckianCalculator::PlanckianPoint(0.2610, 0.2617, 17500),
PlanckianCalculator::PlanckianPoint(0.2609, 0.2615, 17600),
PlanckianCalculator::PlanckianPoint(0.2608, 0.2613, 17700),
PlanckianCalculator::PlanckianPoint(0.2606, 0.2611, 17800),
PlanckianCalculator::PlanckianPoint(0.2605, 0.2609, 17900),
PlanckianCalculator::PlanckianPoint(0.2604, 0.2607, 18000),
PlanckianCalculator::PlanckianPoint(0.2602, 0.2606, 18100),
PlanckianCalculator::PlanckianPoint(0.2601, 0.2604, 18200),
PlanckianCalculator::PlanckianPoint(0.2600, 0.2602, 18300),
PlanckianCalculator::PlanckianPoint(0.2598, 0.2600, 18400),
PlanckianCalculator::PlanckianPoint(0.2597, 0.2598, 18500),
PlanckianCalculator::PlanckianPoint(0.2596, 0.2597, 18600),
PlanckianCalculator::PlanckianPoint(0.2595, 0.2595, 18700),
PlanckianCalculator::PlanckianPoint(0.2593, 0.2593, 18800),
PlanckianCalculator::PlanckianPoint(0.2592, 0.2592, 18900),
PlanckianCalculator::PlanckianPoint(0.2591, 0.2590, 19000),
PlanckianCalculator::PlanckianPoint(0.2590, 0.2588, 19100),
PlanckianCalculator::PlanckianPoint(0.2589, 0.2587, 19200),
PlanckianCalculator::PlanckianPoint(0.2588, 0.2585, 19300),
PlanckianCalculator::PlanckianPoint(0.2587, 0.2584, 19400),
PlanckianCalculator::PlanckianPoint(0.2586, 0.2582, 19500),
PlanckianCalculator::PlanckianPoint(0.2584, 0.2580, 19600),
PlanckianCalculator::PlanckianPoint(0.2583, 0.2579, 19700),
PlanckianCalculator::PlanckianPoint(0.2582, 0.2577, 19800),
PlanckianCalculator::PlanckianPoint(0.2581, 0.2576, 19900),
PlanckianCalculator::PlanckianPoint(0.2580, 0.2574, 20000),
PlanckianCalculator::PlanckianPoint(0.2579, 0.2573, 20100),
PlanckianCalculator::PlanckianPoint(0.2578, 0.2572, 20200),
PlanckianCalculator::PlanckianPoint(0.2577, 0.2570, 20300),
PlanckianCalculator::PlanckianPoint(0.2576, 0.2569, 20400),
PlanckianCalculator::PlanckianPoint(0.2575, 0.2567, 20500),
PlanckianCalculator::PlanckianPoint(0.2574, 0.2566, 20600),
PlanckianCalculator::PlanckianPoint(0.2573, 0.2565, 20700),
PlanckianCalculator::PlanckianPoint(0.2572, 0.2563, 20800),
PlanckianCalculator::PlanckianPoint(0.2571, 0.2562, 20900),
PlanckianCalculator::PlanckianPoint(0.2571, 0.2561, 21000),
PlanckianCalculator::PlanckianPoint(0.2570, 0.2559, 21100),
PlanckianCalculator::PlanckianPoint(0.2569, 0.2558, 21200),
PlanckianCalculator::PlanckianPoint(0.2568, 0.2557, 21300),
PlanckianCalculator::PlanckianPoint(0.2567, 0.2555, 21400),
PlanckianCalculator::PlanckianPoint(0.2566, 0.2554, 21500),
PlanckianCalculator::PlanckianPoint(0.2565, 0.2553, 21600),
PlanckianCalculator::PlanckianPoint(0.2564, 0.2552, 21700),
PlanckianCalculator::PlanckianPoint(0.2564, 0.2550, 21800),
PlanckianCalculator::PlanckianPoint(0.2563, 0.2549, 21900),
PlanckianCalculator::PlanckianPoint(0.2562, 0.2548, 22000),
PlanckianCalculator::PlanckianPoint(0.2561, 0.2547, 22100),
PlanckianCalculator::PlanckianPoint(0.2560, 0.2546, 22200),
PlanckianCalculator::PlanckianPoint(0.2559, 0.2545, 22300),
PlanckianCalculator::PlanckianPoint(0.2559, 0.2543, 22400),
PlanckianCalculator::PlanckianPoint(0.2558, 0.2542, 22500),
PlanckianCalculator::PlanckianPoint(0.2557, 0.2541, 22600),
PlanckianCalculator::PlanckianPoint(0.2556, 0.2540, 22700),
PlanckianCalculator::PlanckianPoint(0.2556, 0.2539, 22800),
PlanckianCalculator::PlanckianPoint(0.2555, 0.2538, 22900),
PlanckianCalculator::PlanckianPoint(0.2554, 0.2537, 23000),
PlanckianCalculator::PlanckianPoint(0.2553, 0.2536, 23100),
PlanckianCalculator::PlanckianPoint(0.2553, 0.2535, 23200),
PlanckianCalculator::PlanckianPoint(0.2552, 0.2534, 23300),
PlanckianCalculator::PlanckianPoint(0.2551, 0.2533, 23400),
PlanckianCalculator::PlanckianPoint(0.2551, 0.2532, 23500),
PlanckianCalculator::PlanckianPoint(0.2550, 0.2531, 23600),
PlanckianCalculator::PlanckianPoint(0.2549, 0.2530, 23700),
PlanckianCalculator::PlanckianPoint(0.2548, 0.2529, 23800),
PlanckianCalculator::PlanckianPoint(0.2548, 0.2528, 23900),
PlanckianCalculator::PlanckianPoint(0.2547, 0.2527, 24000),
PlanckianCalculator::PlanckianPoint(0.2546, 0.2526, 24100),
PlanckianCalculator::PlanckianPoint(0.2546, 0.2525, 24200),
PlanckianCalculator::PlanckianPoint(0.2545, 0.2524, 24300),
PlanckianCalculator::PlanckianPoint(0.2544, 0.2523, 24400),
PlanckianCalculator::PlanckianPoint(0.2544, 0.2522, 24500),
PlanckianCalculator::PlanckianPoint(0.2543, 0.2521, 24600),
PlanckianCalculator::PlanckianPoint(0.2543, 0.2520, 24700),
PlanckianCalculator::PlanckianPoint(0.2542, 0.2519, 24800),
PlanckianCalculator::PlanckianPoint(0.2541, 0.2518, 24900),
PlanckianCalculator::PlanckianPoint(0.2541, 0.2517, 25000),
PlanckianCalculator::PlanckianPoint(0.2540, 0.2516, 25100),
PlanckianCalculator::PlanckianPoint(0.2540, 0.2516, 25200),
PlanckianCalculator::PlanckianPoint(0.2539, 0.2515, 25300),
PlanckianCalculator::PlanckianPoint(0.2538, 0.2514, 25400),
PlanckianCalculator::PlanckianPoint(0.2538, 0.2513, 25500),
PlanckianCalculator::PlanckianPoint(0.2537, 0.2512, 25600),
PlanckianCalculator::PlanckianPoint(0.2537, 0.2511, 25700),
PlanckianCalculator::PlanckianPoint(0.2536, 0.2511, 25800),
PlanckianCalculator::PlanckianPoint(0.2535, 0.2510, 25900),
PlanckianCalculator::PlanckianPoint(0.2535, 0.2509, 26000),
PlanckianCalculator::PlanckianPoint(0.2534, 0.2508, 26100),
PlanckianCalculator::PlanckianPoint(0.2534, 0.2507, 26200),
PlanckianCalculator::PlanckianPoint(0.2533, 0.2507, 26300),
PlanckianCalculator::PlanckianPoint(0.2533, 0.2506, 26400),
PlanckianCalculator::PlanckianPoint(0.2532, 0.2505, 26500),
PlanckianCalculator::PlanckianPoint(0.2532, 0.2504, 26600),
PlanckianCalculator::PlanckianPoint(0.2531, 0.2503, 26700),
PlanckianCalculator::PlanckianPoint(0.2531, 0.2503, 26800),
PlanckianCalculator::PlanckianPoint(0.2530, 0.2502, 26900),
PlanckianCalculator::PlanckianPoint(0.2530, 0.2501, 27000),
PlanckianCalculator::PlanckianPoint(0.2529, 0.2500, 27100),
PlanckianCalculator::PlanckianPoint(0.2529, 0.2500, 27200),
PlanckianCalculator::PlanckianPoint(0.2528, 0.2499, 27300),
PlanckianCalculator::PlanckianPoint(0.2528, 0.2498, 27400),
PlanckianCalculator::PlanckianPoint(0.2527, 0.2497, 27500),
PlanckianCalculator::PlanckianPoint(0.2527, 0.2497, 27600),
PlanckianCalculator::PlanckianPoint(0.2526, 0.2496, 27700),
PlanckianCalculator::PlanckianPoint(0.2526, 0.2495, 27800),
PlanckianCalculator::PlanckianPoint(0.2525, 0.2495, 27900),
PlanckianCalculator::PlanckianPoint(0.2525, 0.2494, 28000),
PlanckianCalculator::PlanckianPoint(0.2524, 0.2493, 28100),
PlanckianCalculator::PlanckianPoint(0.2524, 0.2493, 28200),
PlanckianCalculator::PlanckianPoint(0.2523, 0.2492, 28300),
PlanckianCalculator::PlanckianPoint(0.2523, 0.2491, 28400),
PlanckianCalculator::PlanckianPoint(0.2523, 0.2491, 28500),
PlanckianCalculator::PlanckianPoint(0.2522, 0.2490, 28600),
PlanckianCalculator::PlanckianPoint(0.2522, 0.2489, 28700),
PlanckianCalculator::PlanckianPoint(0.2521, 0.2489, 28800),
PlanckianCalculator::PlanckianPoint(0.2521, 0.2488, 28900),
PlanckianCalculator::PlanckianPoint(0.2520, 0.2487, 29000),
PlanckianCalculator::PlanckianPoint(0.2520, 0.2487, 29100),
PlanckianCalculator::PlanckianPoint(0.2519, 0.2486, 29200),
PlanckianCalculator::PlanckianPoint(0.2519, 0.2485, 29300),
PlanckianCalculator::PlanckianPoint(0.2519, 0.2485, 29400),
PlanckianCalculator::PlanckianPoint(0.2518, 0.2484, 29500),
PlanckianCalculator::PlanckianPoint(0.2518, 0.2484, 29600),
PlanckianCalculator::PlanckianPoint(0.2517, 0.2483, 29700),
PlanckianCalculator::PlanckianPoint(0.2517, 0.2482, 29800),
PlanckianCalculator::PlanckianPoint(0.2517, 0.2482, 29900),
PlanckianCalculator::PlanckianPoint(0.2516, 0.2481, 30000),
PlanckianCalculator::PlanckianPoint(0.2516, 0.2481, 30100),
PlanckianCalculator::PlanckianPoint(0.2515, 0.2480, 30200),
PlanckianCalculator::PlanckianPoint(0.2515, 0.2480, 30300),
PlanckianCalculator::PlanckianPoint(0.2515, 0.2479, 30400),
PlanckianCalculator::PlanckianPoint(0.2514, 0.2478, 30500),
PlanckianCalculator::PlanckianPoint(0.2514, 0.2478, 30600),
PlanckianCalculator::PlanckianPoint(0.2513, 0.2477, 30700),
PlanckianCalculator::PlanckianPoint(0.2513, 0.2477, 30800),
PlanckianCalculator::PlanckianPoint(0.2513, 0.2476, 30900),
PlanckianCalculator::PlanckianPoint(0.2512, 0.2476, 31000),
PlanckianCalculator::PlanckianPoint(0.2512, 0.2475, 31100),
PlanckianCalculator::PlanckianPoint(0.2512, 0.2474, 31200),
PlanckianCalculator::PlanckianPoint(0.2511, 0.2474, 31300),
PlanckianCalculator::PlanckianPoint(0.2511, 0.2473, 31400),
PlanckianCalculator::PlanckianPoint(0.2511, 0.2473, 31500),
PlanckianCalculator::PlanckianPoint(0.2510, 0.2472, 31600),
PlanckianCalculator::PlanckianPoint(0.2510, 0.2472, 31700),
PlanckianCalculator::PlanckianPoint(0.2509, 0.2471, 31800),
PlanckianCalculator::PlanckianPoint(0.2509, 0.2471, 31900),
PlanckianCalculator::PlanckianPoint(0.2509, 0.2470, 32000),
PlanckianCalculator::PlanckianPoint(0.2508, 0.2470, 32100),
PlanckianCalculator::PlanckianPoint(0.2508, 0.2469, 32200),
PlanckianCalculator::PlanckianPoint(0.2508, 0.2469, 32300),
PlanckianCalculator::PlanckianPoint(0.2507, 0.2468, 32400),
PlanckianCalculator::PlanckianPoint(0.2507, 0.2468, 32500),
PlanckianCalculator::PlanckianPoint(0.2507, 0.2467, 32600),
PlanckianCalculator::PlanckianPoint(0.2506, 0.2467, 32700),
PlanckianCalculator::PlanckianPoint(0.2506, 0.2466, 32800),
PlanckianCalculator::PlanckianPoint(0.2506, 0.2466, 32900),
PlanckianCalculator::PlanckianPoint(0.2505, 0.2465, 33000),
PlanckianCalculator::PlanckianPoint(0.2505, 0.2465, 33100),
PlanckianCalculator::PlanckianPoint(0.2505, 0.2464, 33200),
PlanckianCalculator::PlanckianPoint(0.2505, 0.2464, 33300),
PlanckianCalculator::PlanckianPoint(0.2504, 0.2463, 33400),
PlanckianCalculator::PlanckianPoint(0.2504, 0.2463, 33500),
PlanckianCalculator::PlanckianPoint(0.2504, 0.2463, 33600),
PlanckianCalculator::PlanckianPoint(0.2503, 0.2462, 33700),
PlanckianCalculator::PlanckianPoint(0.2503, 0.2462, 33800),
PlanckianCalculator::PlanckianPoint(0.2503, 0.2461, 33900),
PlanckianCalculator::PlanckianPoint(0.2502, 0.2461, 34000),
PlanckianCalculator::PlanckianPoint(0.2502, 0.2460, 34100),
PlanckianCalculator::PlanckianPoint(0.2502, 0.2460, 34200),
PlanckianCalculator::PlanckianPoint(0.2502, 0.2459, 34300),
PlanckianCalculator::PlanckianPoint(0.2501, 0.2459, 34400),
PlanckianCalculator::PlanckianPoint(0.2501, 0.2459, 34500),
PlanckianCalculator::PlanckianPoint(0.2501, 0.2458, 34600),
PlanckianCalculator::PlanckianPoint(0.2500, 0.2458, 34700),
PlanckianCalculator::PlanckianPoint(0.2500, 0.2457, 34800),
PlanckianCalculator::PlanckianPoint(0.2500, 0.2457, 34900),
PlanckianCalculator::PlanckianPoint(0.2500, 0.2456, 35000),
PlanckianCalculator::PlanckianPoint(0.2499, 0.2456, 35100),
PlanckianCalculator::PlanckianPoint(0.2499, 0.2456, 35200),
PlanckianCalculator::PlanckianPoint(0.2499, 0.2455, 35300),
PlanckianCalculator::PlanckianPoint(0.2498, 0.2455, 35400),
PlanckianCalculator::PlanckianPoint(0.2498, 0.2454, 35500),
PlanckianCalculator::PlanckianPoint(0.2498, 0.2454, 35600),
PlanckianCalculator::PlanckianPoint(0.2498, 0.2454, 35700),
PlanckianCalculator::PlanckianPoint(0.2497, 0.2453, 35800),
PlanckianCalculator::PlanckianPoint(0.2497, 0.2453, 35900),
PlanckianCalculator::PlanckianPoint(0.2497, 0.2452, 36000),
PlanckianCalculator::PlanckianPoint(0.2497, 0.2452, 36100),
PlanckianCalculator::PlanckianPoint(0.2496, 0.2452, 36200),
PlanckianCalculator::PlanckianPoint(0.2496, 0.2451, 36300),
PlanckianCalculator::PlanckianPoint(0.2496, 0.2451, 36400),
PlanckianCalculator::PlanckianPoint(0.2496, 0.2450, 36500),
PlanckianCalculator::PlanckianPoint(0.2495, 0.2450, 36600),
PlanckianCalculator::PlanckianPoint(0.2495, 0.2450, 36700),
PlanckianCalculator::PlanckianPoint(0.2495, 0.2449, 36800),
PlanckianCalculator::PlanckianPoint(0.2495, 0.2449, 36900),
PlanckianCalculator::PlanckianPoint(0.2494, 0.2449, 37000),
PlanckianCalculator::PlanckianPoint(0.2494, 0.2448, 37100),
PlanckianCalculator::PlanckianPoint(0.2494, 0.2448, 37200),
PlanckianCalculator::PlanckianPoint(0.2494, 0.2447, 37300),
PlanckianCalculator::PlanckianPoint(0.2493, 0.2447, 37400),
PlanckianCalculator::PlanckianPoint(0.2493, 0.2447, 37500),
PlanckianCalculator::PlanckianPoint(0.2493, 0.2446, 37600),
PlanckianCalculator::PlanckianPoint(0.2493, 0.2446, 37700),
PlanckianCalculator::PlanckianPoint(0.2492, 0.2446, 37800),
PlanckianCalculator::PlanckianPoint(0.2492, 0.2445, 37900),
PlanckianCalculator::PlanckianPoint(0.2492, 0.2445, 38000),
PlanckianCalculator::PlanckianPoint(0.2492, 0.2445, 38100),
PlanckianCalculator::PlanckianPoint(0.2491, 0.2444, 38200),
PlanckianCalculator::PlanckianPoint(0.2491, 0.2444, 38300),
PlanckianCalculator::PlanckianPoint(0.2491, 0.2444, 38400),
PlanckianCalculator::PlanckianPoint(0.2491, 0.2443, 38500),
PlanckianCalculator::PlanckianPoint(0.2491, 0.2443, 38600),
PlanckianCalculator::PlanckianPoint(0.2490, 0.2443, 38700),
PlanckianCalculator::PlanckianPoint(0.2490, 0.2442, 38800),
PlanckianCalculator::PlanckianPoint(0.2490, 0.2442, 38900),
PlanckianCalculator::PlanckianPoint(0.2490, 0.2442, 39000),
PlanckianCalculator::PlanckianPoint(0.2489, 0.2441, 39100),
PlanckianCalculator::PlanckianPoint(0.2489, 0.2441, 39200),
PlanckianCalculator::PlanckianPoint(0.2489, 0.2441, 39300),
PlanckianCalculator::PlanckianPoint(0.2489, 0.2440, 39400),
PlanckianCalculator::PlanckianPoint(0.2489, 0.2440, 39500),
PlanckianCalculator::PlanckianPoint(0.2488, 0.2440, 39600),
PlanckianCalculator::PlanckianPoint(0.2488, 0.2439, 39700),
PlanckianCalculator::PlanckianPoint(0.2488, 0.2439, 39800),
PlanckianCalculator::PlanckianPoint(0.2488, 0.2439, 39900),
PlanckianCalculator::PlanckianPoint(0.2487, 0.2438, 40000),
PlanckianCalculator::PlanckianPoint(0.2487, 0.2438, 40100),
PlanckianCalculator::PlanckianPoint(0.2487, 0.2438, 40200),
PlanckianCalculator::PlanckianPoint(0.2487, 0.2437, 40300),
PlanckianCalculator::PlanckianPoint(0.2487, 0.2437, 40400),
PlanckianCalculator::PlanckianPoint(0.2486, 0.2437, 40500),
PlanckianCalculator::PlanckianPoint(0.2486, 0.2437, 40600),
PlanckianCalculator::PlanckianPoint(0.2486, 0.2436, 40700),
PlanckianCalculator::PlanckianPoint(0.2486, 0.2436, 40800),
PlanckianCalculator::PlanckianPoint(0.2486, 0.2436, 40900),
PlanckianCalculator::PlanckianPoint(0.2485, 0.2435, 41000),
PlanckianCalculator::PlanckianPoint(0.2485, 0.2435, 41100),
PlanckianCalculator::PlanckianPoint(0.2485, 0.2435, 41200),
PlanckianCalculator::PlanckianPoint(0.2485, 0.2434, 41300),
PlanckianCalculator::PlanckianPoint(0.2485, 0.2434, 41400),
PlanckianCalculator::PlanckianPoint(0.2485, 0.2434, 41500),
PlanckianCalculator::PlanckianPoint(0.2484, 0.2434, 41600),
PlanckianCalculator::PlanckianPoint(0.2484, 0.2433, 41700),
PlanckianCalculator::PlanckianPoint(0.2484, 0.2433, 41800),
PlanckianCalculator::PlanckianPoint(0.2484, 0.2433, 41900),
PlanckianCalculator::PlanckianPoint(0.2484, 0.2432, 42000),
PlanckianCalculator::PlanckianPoint(0.2483, 0.2432, 42100),
PlanckianCalculator::PlanckianPoint(0.2483, 0.2432, 42200),
PlanckianCalculator::PlanckianPoint(0.2483, 0.2432, 42300),
PlanckianCalculator::PlanckianPoint(0.2483, 0.2431, 42400),
PlanckianCalculator::PlanckianPoint(0.2483, 0.2431, 42500),
PlanckianCalculator::PlanckianPoint(0.2482, 0.2431, 42600),
PlanckianCalculator::PlanckianPoint(0.2482, 0.2431, 42700),
PlanckianCalculator::PlanckianPoint(0.2482, 0.2430, 42800),
PlanckianCalculator::PlanckianPoint(0.2482, 0.2430, 42900),
PlanckianCalculator::PlanckianPoint(0.2482, 0.2430, 43000),
PlanckianCalculator::PlanckianPoint(0.2482, 0.2429, 43100),
PlanckianCalculator::PlanckianPoint(0.2481, 0.2429, 43200),
PlanckianCalculator::PlanckianPoint(0.2481, 0.2429, 43300),
PlanckianCalculator::PlanckianPoint(0.2481, 0.2429, 43400),
PlanckianCalculator::PlanckianPoint(0.2481, 0.2428, 43500),
PlanckianCalculator::PlanckianPoint(0.2481, 0.2428, 43600),
PlanckianCalculator::PlanckianPoint(0.2481, 0.2428, 43700),
PlanckianCalculator::PlanckianPoint(0.2480, 0.2428, 43800),
PlanckianCalculator::PlanckianPoint(0.2480, 0.2427, 43900),
PlanckianCalculator::PlanckianPoint(0.2480, 0.2427, 44000),
PlanckianCalculator::PlanckianPoint(0.2480, 0.2427, 44100),
PlanckianCalculator::PlanckianPoint(0.2480, 0.2427, 44200),
PlanckianCalculator::PlanckianPoint(0.2480, 0.2426, 44300),
PlanckianCalculator::PlanckianPoint(0.2479, 0.2426, 44400),
PlanckianCalculator::PlanckianPoint(0.2479, 0.2426, 44500),
PlanckianCalculator::PlanckianPoint(0.2479, 0.2426, 44600),
PlanckianCalculator::PlanckianPoint(0.2479, 0.2425, 44700),
PlanckianCalculator::PlanckianPoint(0.2479, 0.2425, 44800),
PlanckianCalculator::PlanckianPoint(0.2479, 0.2425, 44900),
PlanckianCalculator::PlanckianPoint(0.2478, 0.2425, 45000),
PlanckianCalculator::PlanckianPoint(0, 0, 0)};
//-----------------------------------------------------------------------------
PlanckianCalculator::PlanckianPoint PlanckianCalculator::ProjectOnPlanckian(const Chromacity& ch)
  {
  PlanckianPoint pnt(0,0,0);
  float min_dist = 1e20;

  for (ulong i = 0; m_planckian_points[i].m_t != 0; ++i)
    {
    // Build the line approx at current segment:
    /*
    float a = (m_planckian_points[i+1][1] - m_planckian_points[i][1])/
              (m_planckian_points[i+1][0] - m_planckian_points[i][0]),
          b = m_planckian_points[i][1] - a*m_planckian_points[i][0],

          x = (m_planckian_points[i][0] - a*b + a*m_planckian_points[i][1])/(1+a*a);
    if (x < m_planckian_points[i+1][0]
        || x > m_planckian_points[i][0])
      y = std::min(sqrt(sqr(ch.m_red - m_planckian_points[i][0]
    */
    float dist = sqrt(sqr(ch.m_r - m_planckian_points[i].m_chroma.m_r)
                      + sqr(ch.m_g - m_planckian_points[i].m_chroma.m_g));

    if (dist < min_dist)
      {
      min_dist = dist;
      pnt = m_planckian_points[i];
      }
    }

  return pnt;
  }
//-----------------------------------------------------------------------------
PlanckianCalculator::PlanckianPoint PlanckianCalculator::ProjectOnPlanckian(float temp)
  {
  cmsCIExyY wpnt;

  if (!cmsWhitePointFromTemp(temp, &wpnt))
    {
    mg_log << WARN << "Failed to obtain chromacity of white point with temperature " << temp << endlog;
    mg_log << WARN << "Will use lookup table instead" << endlog;

    for (ulong i = 0; m_planckian_points[i].m_t != 0; ++i)
      if (fabs(m_planckian_points[i].m_t - temp) <= 50)
        return m_planckian_points[i];
    }
  else
    return PlanckianPoint(wpnt.x, wpnt.y, temp);

  assert(0);

  return PlanckianPoint(0, 0, 0);
  }
//-----------------------------------------------------------------------------
WBCalculator::WBCalculator(): IParametrizable("WBCalculator")
  {
  m_zone_width = 2;
  //m_zone_threshold_dev = 0.1;
  m_zone_threshold_dev = 0.15;
  m_threshold_dark = 16;

  m_threshold_clamp = 256;
  m_threshold_clamp2 = 512;
  m_threshold_clamp3 = 1024;
  m_fl_use_planckian = true;
  //m_fl_use_planckian = false;
  }
//-----------------------------------------------------------------------------
void WBCalculator::GetParameters(Params& o_params) const
  {
  }
//-----------------------------------------------------------------------------
void WBCalculator::GetDescriptions(Params& o_params) const
  {
  }
//-----------------------------------------------------------------------------
void WBCalculator::SetParameters(Params& i_params)
  {
  }
//-----------------------------------------------------------------------------
bool WBCalculator::IsClamped(const Image& image, unsigned row,
                             unsigned col, const RawImageInfo& image_info)
  {
  for (unsigned r = row - m_zone_width; r <= row + m_zone_width; ++r)
    for (unsigned c = col - m_zone_width; c <= col + m_zone_width; ++c)
      {
      uchar ind = image_info.GetCurrentColorIndex(r, c);

      // I don't know why, but it seems that values may be clamped
      // even if they didn't reach the maximum possible value
      // (maybe the result of ISO boost?)
      // So, to check for clamping we have to use not theoretically 
      // maximum value, but maximum value that occured in the image
      if ((int)image.At(r,c)[ind] + (int)image_info.GetNoise(ind) >= (int)image_info.GetMaxColor(ind) - m_threshold_clamp)
        return true;

      int v_r = (int)image.At(r,c)[0] + (int)image_info.GetNoise(0),
          v_g = (int)image.At(r,c)[1] + (int)image_info.GetNoise(1),
          v_b = (int)image.At(r,c)[2] + (int)image_info.GetNoise(2),
          c_r = (int)image_info.GetMaxColor(0) - m_threshold_clamp2,
          c_g = (int)image_info.GetMaxColor(1) - m_threshold_clamp2,
          c_b = (int)image_info.GetMaxColor(2) - m_threshold_clamp2;
      
      if (v_r >= c_r && v_g >= c_g ||
          v_b >= c_b && v_g >= c_g ||
          v_r >= c_r && v_b >= c_b)
        return true;
      
      c_r = (int)image_info.GetMaxColor(0) - m_threshold_clamp3;
      c_g = (int)image_info.GetMaxColor(1) - m_threshold_clamp3;
      c_b = (int)image_info.GetMaxColor(2) - m_threshold_clamp3;
      
      if (v_r >= c_r && v_g >= c_g && v_b >= c_b)
        return true;
      }

  return false;
  }
//-----------------------------------------------------------------------------
bool WBCalculator::IsDeviated(const Image& image, unsigned row,
                              unsigned col, const RawImageInfo& image_info,
                              float threshold, float* p_deviation)
  {
  if (threshold == 0)
      threshold = m_zone_threshold_dev;

  ulong size = (2*m_zone_width+1)*(2*m_zone_width+1);

  std::vector<float> hues_r(size), hues_b(size);

  ulong count = 0;

  float sum_hue_r = 0, sum_hue_b = 0;

  for (unsigned r = row - m_zone_width; r <= row + m_zone_width; ++r)
    for (unsigned c = col - m_zone_width; c <= col + m_zone_width; ++c)
      {
      if (!image.At(r,c)[1])
        continue;

      float hr = (float)image.At(r,c)[0]/image.At(r,c)[1],
            hb = (float)image.At(r,c)[2]/image.At(r,c)[1];
      sum_hue_r += hr;
      sum_hue_b += hb;
      hues_r[count] = hr;
      hues_b[count] = hb;
      ++count;
      }

  if (count)
    {
    sum_hue_r /= count;
    sum_hue_b /= count;
    }
  else
    return true;

  for (unsigned i = 0; i < count; ++i)
    {
    if (p_deviation)
      *p_deviation += sqr(hues_r[i] - sum_hue_r) + sqr(hues_b[i] - sum_hue_b);

    if (fabs(hues_r[i] - sum_hue_r) > threshold ||
        fabs(hues_b[i] - sum_hue_b) > threshold)
      return true;
    }

  return false;
  }
//-----------------------------------------------------------------------------
void WBCalculator::GetTheBestSeed(const Mask& mask,
                                  Position& pos, const Image& image,
                                  const RawImageInfo& image_info)
  {
  mg_log << TRACE << "Selecting the best seed to start segment" << endlog;
  unsigned tmp_width = m_zone_width;
  m_zone_width = 4;
  float tmp_dev = m_zone_threshold_dev;
  m_zone_threshold_dev = 1e20;

  ulong dark = 0, clamp = 0, dev = 0;

  float min_dev = 1e20;
  
  for (unsigned row = m_zone_width + image.GetTrimWidth() + 2; row < image.GetHeight() - m_zone_width - image.GetTrimWidth() - 2; ++row)
    for (unsigned col = m_zone_width + image.GetTrimWidth() + 2; col < image.GetWidth() - m_zone_width - image.GetTrimWidth() - 2; ++col)
      {
      if (mask.Get(row, col))
        continue;

      if (image.GetIntensity(row, col) < m_threshold_dark)
        {
        ++dark;
        continue;
        }

      if (IsClamped(image, row, col, image_info))
        {
        ++clamp;
        continue;
        }

      float deviation;
      if (IsDeviated(image, row, col, image_info, m_zone_threshold_dev, &deviation))
        {
        ++dev;
        continue;
        }

      if (deviation < min_dev)
        {
        min_dev = deviation;
        pos = Position(row, col);
        }
      }

  mg_log << TRACE << "In the BestSeed; Dark blocks: " << dark << ", clamped blocks: " << clamp
    << ", deviated blocks:" << dev << endlog;
  mg_log << TRACE << "Seed point: " << pos.m_row << ", " << pos.m_col << endlog;

  m_zone_width = tmp_width;
  m_zone_threshold_dev = tmp_dev;
  }
//-----------------------------------------------------------------------------
bool WBCalculator::IsGoodForSegment(const Mask& mask, const Image& image,
                          const RawImageInfo& image_info,
                          const Position& start_pos,
                          const Position& pos_to_test)
  {
  if (mask.Get(pos_to_test.m_row, pos_to_test.m_col))
    return false;

  if (image.GetIntensity(pos_to_test.m_row, pos_to_test.m_col) < m_threshold_dark)
    return false;

  if (IsClamped(image, pos_to_test.m_row, pos_to_test.m_col, image_info))
    return false;

  //mg_log << DEBUG << "Distance: " << start_pos.Distance(pos_to_test) << ", coef: " <<
  //  1 + pow(start_pos.Distance(pos_to_test), 0.25) << endlog;
  //if (IsDeviated(image, pos_to_test.m_row, pos_to_test.m_col, image_info,
  //    m_zone_threshold_dev/pow(start_pos.Distance(pos_to_test), 0.25)))
  if (IsDeviated(image, pos_to_test.m_row, pos_to_test.m_col, image_info,
      m_zone_threshold_dev/(1+log((double)(start_pos.Distance(pos_to_test))/5))))
    return false;

  return true;
  }
//-----------------------------------------------------------------------------
void WBCalculator::SegmentImage(WBSegmentPixelNotifier* p_notifier,
                                const Image& image, const RawImageInfo& image_info,
                                ulong max_segments_num)
  {
  mg_log << TRACE << "Segmenting image" << endlog;

#ifdef WRITE_SEGMENTED_IMAGE
  Image img_res(image.GetWidth(), image.GetHeight());
  img_res.Zero();
#endif

  Mask mask(image.GetWidth(), image.GetHeight());

  for (ulong cur_segm = 0; cur_segm < max_segments_num; ++cur_segm)
    {
    // We should find the best pixel to start growth of segment
    // For now, let's search for the pixel with smallest deviation
    Position pos(0, 0);

    GetTheBestSeed(mask, pos, image, image_info);

    std::deque<Position> queue;
    queue.push_back(pos);
    p_notifier -> NewSegment(cur_segm);
    p_notifier -> NewPixelInSegment(cur_segm, pos);
    mask.Set(pos.m_row, pos.m_col, 1);

    while(!queue.empty())
      {
      Position cur_pos = queue.front();
      queue.pop_front();

      for (unsigned r = cur_pos.m_row - 1; r <= cur_pos.m_row + 1
           && r >= m_zone_width + image.GetTrimWidth() + 2
           && r < image.GetHeight() - m_zone_width - image.GetTrimWidth() - 2; ++r)
        for (unsigned c = cur_pos.m_col - 1; c <= cur_pos.m_col + 1
             && c >= m_zone_width + image.GetTrimWidth() + 2
             && c < image.GetWidth() - m_zone_width - image.GetTrimWidth() - 2; ++c)
          {
          Position tmp_pos(r,c);

          if (tmp_pos == cur_pos)
            continue;

//          if (tmp_pos.Distance(pos) < cur_pos.Distance(pos))
//            continue;

          if (IsGoodForSegment(mask, image, image_info, pos, tmp_pos))
            {
            p_notifier -> NewPixelInSegment(cur_segm, tmp_pos);
            queue.push_back(tmp_pos);
            mask.Set(tmp_pos.m_row, tmp_pos.m_col, 1);

#ifdef WRITE_SEGMENTED_IMAGE
            img_res.At(tmp_pos) = image.At(tmp_pos);
#endif
            }
          }
      }
    }

#ifdef WRITE_SEGMENTED_IMAGE
  ImageTransfer::GetTransfer().Write(WRITE_SEGMENTED_IMAGE, img_res);
#endif
  }
//-----------------------------------------------------------------------------