/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * WBGreyWorldCalculator - white balance calculation based on "Grey World"
 * assumption.
******************************************************************************/

#include "stdafx.h"

#include "WBGreyWorldCalculator.h"
#include "RawImageInfo.h"
//-----------------------------------------------------------------------------
WBGreyWorldCalculator& WBGreyWorldCalculator::GetWBCalculator()
  {
  static WBGreyWorldCalculator calculator;
  return calculator;
  }
//-----------------------------------------------------------------------------
WBGreyWorldCalculator::WBGreyWorldCalculator(): IParametrizable("WBGreyWorldCalculator")
  {
  }
//-----------------------------------------------------------------------------
void WBGreyWorldCalculator::GetParameters(Params& o_params) const
  {
  }
//-----------------------------------------------------------------------------
void WBGreyWorldCalculator::GetDescriptions(Params& o_params) const
  {
  }
//-----------------------------------------------------------------------------
void WBGreyWorldCalculator::SetParameters(Params& i_params)
  {
  }
//-----------------------------------------------------------------------------
WhiteBalance WBGreyWorldCalculator::CalculateWB(const Image& image, const RawImageInfo& image_info)
  {
  mg_log << INFO << "Running WB GreyWorld Calculator" << endlog;

  double val[3] = {0, 0, 0};

  for (unsigned row = image.GetTrimWidth(); row < image.GetHeight() - image.GetTrimWidth(); ++row)
    for (unsigned col = image.GetTrimWidth(); col < image.GetHeight() - image.GetTrimWidth(); ++col)
      {
      uchar ind = image_info.GetCurrentColorIndex(row, col);
      val[ind] += image.At(row, col)[ind];
      }

  ulong size = (image.GetHeight() - 2*image.GetTrimWidth())*
    (image.GetWidth() - 2*image.GetTrimWidth());

  val[0] /= size/4;
  val[1] /= size/2;
  val[2] /= size/4;

  WhiteBalance res;

  if (m_fl_use_planckian)
    {
    Chromacity ch(val[0]/(val[0] + val[1] + val[2]),
                  val[1]/(val[0] + val[1] + val[2]));

    mg_log << DEBUG << "Projecting chromacity: " << ch.m_r
      << ", " << ch.m_g << endlog;

    PlanckianPoint pnt =
      ProjectOnPlanckian(ch);
    mg_log << DEBUG << "Planckian projection: " << pnt.m_chroma.m_r
      << ", " << pnt.m_chroma.m_g << ", " << pnt.m_t << endlog;

    res.m_red = pnt.m_chroma.m_g/pnt.m_chroma.m_r;
    res.m_green = 1.0;
    res.m_blue = pnt.m_chroma.m_g/(1 - pnt.m_chroma.m_r - pnt.m_chroma.m_g);
    }
  else
    {
    res.m_red = val[1]/val[0];
    res.m_green = 1.0;
    res.m_blue = val[1]/val[2];
    }

  return res;
  }
//-----------------------------------------------------------------------------