/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * WBScaleByMaxCalculator - white balance calculation based on "white patch"
 * assumption.
******************************************************************************/

#include "stdafx.h"

#include "WBScaleByMaxCalculator.h"

#define DUMP_MAX_VALS
//-----------------------------------------------------------------------------
WBScaleByMaxCalculator& WBScaleByMaxCalculator::GetWBCalculator()
  {
  static WBScaleByMaxCalculator calculator;
  return calculator;
  }
//-----------------------------------------------------------------------------
WBScaleByMaxCalculator::WBScaleByMaxCalculator(): IParametrizable("WBScaleByMaxCalculator")
  {
  }
//-----------------------------------------------------------------------------
void WBScaleByMaxCalculator::GetParameters(Params& o_params) const
  {
  }
//-----------------------------------------------------------------------------
void WBScaleByMaxCalculator::GetDescriptions(Params& o_params) const
  {
  }
//-----------------------------------------------------------------------------
void WBScaleByMaxCalculator::SetParameters(Params& i_params)
  {
  }
//-----------------------------------------------------------------------------
WhiteBalance WBScaleByMaxCalculator::CalculateWB(const Image& image, const RawImageInfo& image_info)
  {
  mg_log << INFO << "Running WB ScaleByMax Calculator" << endlog;

  ulong m_avg_size = 5;
  std::set<ushort> max_c[3];

  for (unsigned row = image.GetTrimWidth(); row < image.GetHeight() - image.GetTrimWidth(); ++row)
    for (unsigned col = image.GetTrimWidth(); col < image.GetHeight() - image.GetTrimWidth(); ++col)
      {
      if (IsClamped(image, row, col, image_info)
          || IsDeviated(image, row, col, image_info))
        continue;

      for (uchar ind = 0; ind < 3; ++ind)
        {
        ushort v = image.At(row, col)[ind];

        if (max_c[ind].size() < m_avg_size)
          max_c[ind].insert(v);
        else if (v > *max_c[ind].begin())
          {
          max_c[ind].erase(max_c[ind].begin());
          max_c[ind].insert(v);
          }
        }
      }

  double val[3] = {0, 0, 0};

  for (uchar ind = 0; ind < 3; ++ind)
    for (std::set<ushort>::const_iterator it = max_c[ind].begin();
         it != max_c[ind].end(); ++it)
      val[ind] += *it;

  WhiteBalance res;
  res.m_red = val[1]/val[0];
  res.m_green = 1.0;
  res.m_blue = val[1]/val[2];

  return res;
  }
//-----------------------------------------------------------------------------