/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef WEIGHTED_EDGE_BASE_H
#define WEIGHTED_EDGE_BASE_H

#include "Image.h"
#include "RawImageInfo.h"
//-----------------------------------------------------------------------------
class WeightedEdgeBase
  {
  public:
    float GetD(const Image& image, unsigned row, unsigned col, uchar dir);

    float GetEdgeWeight(const Image& image, unsigned row, unsigned col, uchar dir);

  private:
    static const float sqr2, sqr22;
    RawImageInfo* mp_image_info;
  };
//-----------------------------------------------------------------------------
// 1 2 3
// 4 5 6
// 7 8 9
//
//-----------------------------------------------------------------------------
inline float WeightedEdgeBase::GetD(const Image& image, unsigned row, unsigned col, uchar dir)
  {
  uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
  uchar ind1 = mp_image_info -> GetCurrentColorIndex(row - 1, col);
  uchar ind2 = mp_image_info -> GetCurrentColorIndex(row - 1, col - 1);

  switch (dir)
    {
    case 1:
    case 9:
      if (ind == 1)
        {
        float g1 = fabs(((int)image.At(row - 1, col - 1)[1] - (int)image.At(row, col)[1]) / sqr2);
        float g2 = fabs(((int)image.At(row + 1, col + 1)[1] - (int)image.At(row, col)[1]) / sqr2);

        return std::max(g1, g2);
        }
      else
        return ((int)image.At(row - 1, col - 1)[ind2] - (int)image.At(row + 1, col + 1)[ind2]) / sqr22;

    case 2:
    case 8:
      return ((int)image.At(row - 1, col)[ind1] - (int)image.At(row + 1, col)[ind1]) / 2.0;

    case 3:
    case 7:
      if (ind == 1)
        {
        float g1 = fabs(((int)image.At(row - 1, col + 1)[1] - (int)image.At(row, col)[1]) / sqr2);
        float g2 = fabs(((int)image.At(row + 1, col - 1)[1] - (int)image.At(row, col)[1]) / sqr2);
        return std::max(g1, g2);
        }
      else
        return ((int)image.At(row - 1, col + 1)[ind2] - (int)image.At(row + 1, col - 1)[ind2]) / sqr22;

    case 4:
    case 6:
      return ((int)image.At(row, col - 1)[ind1] - (int)image.At(row, col + 1)[ind1]) / 2.0;
    }

  return 0.0; // To avoid warnings
  }
//-----------------------------------------------------------------------------
inline float WeightedEdgeBase::GetEdgeWeight(const Image& image, unsigned row, unsigned col, uchar dir)
  {
  float d1, d2;

  switch (dir)
    {
    case 1:
      d1 = GetD(image, row, col, dir);
      d2 = GetD(image, row - 1, col - 1, dir);
      break;

    case 2:
      d1 = GetD(image, row, col, dir);
      d2 = GetD(image, row - 1, col, dir);
      break;

    case 3:
      d1 = GetD(image, row, col, dir);
      d2 = GetD(image, row - 1, col + 1, dir);
      break;

    case 4:
      d1 = GetD(image, row, col, dir);
      d2 = GetD(image, row, col - 1, dir);
      break;

    case 6:
      d1 = GetD(image, row, col, dir);
      d2 = GetD(image, row, col + 1, dir);
      break;

    case 7:
      d1 = GetD(image, row, col, dir);
      d2 = GetD(image, row + 1, col - 1, dir);
      break;

    case 8:
      d1 = GetD(image, row, col, dir);
      d2 = GetD(image, row + 1, col, dir);
      break;

    case 9:
      d1 = GetD(image, row, col, dir);
      d2 = GetD(image, row + 1, col + 1, dir);
      break;
    }

  //return 1/sqrt(1 + d1*d1 + d2*d2);
  return 1/(1 + fabs(d1) + fabs(d2));
  }
//-----------------------------------------------------------------------------

#endif // WEIGHTED_EDGE_BASE_H
