/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef WEIGHTED_EDGE_IMAGE_PROCESSOR_H
#define WEIGHTED_EDGE_IMAGE_PROCESSOR_H

#include "WeightedEdgeBase.h"
#include "ImageProcessor.h"
//-----------------------------------------------------------------------------
class WeightedEdgeImageProcessor: public ImageProcessor, private WeightedEdgeBase
  {
  public:
    static WeightedEdgeImageProcessor& GetImageProcessor();

    void Process(Image& image, const RawImageInfo& image_info, const std::string& method);

    void CorrectHue(Image& image, const RawImageInfo& image_info);

    void ApplyDiffusionFlow(Image& image, const RawImageInfo& image_info);

    virtual void GetParameters(Params& o_params) const;

    virtual void GetDescriptions(Params& o_params) const;

    virtual void SetParameters(Params& i_params);

  protected:
    WeightedEdgeImageProcessor();

  private:
    typedef ImageTmpl<PixelTmpl<float, 3> > MatrixG;
    typedef ImageTmpl<float> Gaussian;
    const RawImageInfo* mp_image_info;
    float m_alpha, m_amount, m_radius, m_smooth_coef;
    ulong m_diff_num_of_iter, m_hue_num_of_iter;
    ushort m_small_threshold;
    bool m_fl_variable_alpha;

  private:
    float CorrectGreenByChroma(Image& image, unsigned row, unsigned col, uchar ind, float coefs[]);

    float CorrectGreen(Image& image, unsigned row, unsigned col, float coefs[]);

    float CorrectChroma(Image& image, unsigned row, unsigned col, uchar ind, float coefs[]);

    // Diffusion section:
    void ComputeG(const Image& img_src, unsigned row, unsigned col,
                  MatrixG& g);

    void CreateGaussian(Gaussian& gs, float r);

    void DiffuseG(const Image& img, const MatrixG& g_src, MatrixG& g_dest,
                  unsigned row, unsigned col, const Gaussian& gs);

    void ApplyStep(Image& img, unsigned row, unsigned col,
                   const MatrixG& g, float step);
  };
//-----------------------------------------------------------------------------

#endif // WEIGHTED_EDGE_IMAGE_PROCESSOR_H
