/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * logstream - logger class that supports multiple-destination logging,
 * per-destination configuration, etc.
******************************************************************************/

#include "stdafx.h"
#include "logstream.h"

#include <iostream>
#include <typeinfo>

//--------------------------------------------------------------------
LogErrorLevel CRIT(0), ERR(1), WARN(2), INFO(3), TRACE(4), DEBUG(5);
//--------------------------------------------------------------------
std::string logstream::mg_level_str[6] = { "CRIT", "ERR", "WARN", "INFO", "TRACE", "DEBUG"};
//--------------------------------------------------------------------
logstream mg_log;
//--------------------------------------------------------------------
logstream::logstream(): m_curr_level(INFO)
  {
  SetOutput(std::clog);
  m_max_history_lines = 100;
  }
//--------------------------------------------------------------------
logstream::~logstream()
  {
  *this << TRACE;

  for (std::list<logstream::LogInfo>::iterator it = m_outs.begin(); it != m_outs.end(); ++it)
    {
    if (it -> m_fl_is_responsible) // otherwise stream might be destroyed already
      {
      *it -> mp_out << FormatMessage("Logging is finished", it -> m_fl_add_time, it -> m_fl_add_level) << std::endl;
      it -> mp_out -> flush();
      }
    }
  }
//--------------------------------------------------------------------
//logstream& logstream::operator << (LogErrorLevel i_lev)
std::ostream& logstream::operator << (LogErrorLevel i_lev)
  {
  m_curr_level = i_lev;
  struct tm *newtime;
  time_t long_time;

  time( &long_time );
  newtime = localtime( &long_time );
  m_curr_time.assign(asctime(newtime));

  if (*(m_curr_time.end()-1) == '\n')
    m_curr_time.erase(m_curr_time.end() - 1);

  return *this;
  }
//--------------------------------------------------------------------
void logstream::PrintAllRecentLogMessages()
  {
  *this << CRIT;

  for (std::list<LogInfo>::iterator it = m_outs.begin(); it != m_outs.end(); ++it)
    {
    *it -> mp_out << std::endl;
    *it -> mp_out << FormatMessage("Printing all recent messages:", true, true) << std::endl;

    if (!m_max_history_lines)
      *it -> mp_out << FormatMessage("History is disabled - cannot print recent messages", true, true) << std::endl;
    else
      {
      for (std::list<std::string>::iterator it1 = m_history.begin();
           it1 != m_history.end(); ++it1)
        *it -> mp_out << *it1 << std::endl;

      *this << CRIT;
      *it -> mp_out << FormatMessage("End of recent messages", true, true) << std::endl;
      }
    }
  }
//--------------------------------------------------------------------
std::string logstream::FormatMessage(const std::string& i_str, bool fl_add_time,
                                     bool fl_add_level)
  {
  std::string out;

  if (fl_add_time)
    {
    out = m_curr_time;
    out += " ";
    }

  if (fl_add_level)
    {
    out += mg_level_str[m_curr_level.GetErrorLevel()];
    out += ": ";
    }

  out += i_str;

  return out;
  }
//--------------------------------------------------------------------
std::ostream& endlog(std::ostream& os)
  {
  try
    {
    logstream& ls = dynamic_cast<logstream&>(os);
    }
  catch (std::bad_cast)
    {
    os << "Cannot convert stream to logstream" << std::endl;
    throw;
    }

  logstream& ls = dynamic_cast<logstream&>(os);

  if (ls.m_max_history_lines)
    {
    while (ls.m_history.size() >= ls.m_max_history_lines)
      ls.m_history.erase(ls.m_history.begin());

    ls.m_history.push_back(ls.FormatMessage(ls.str(), true, true));
    }

  for (std::list<logstream::LogInfo>::iterator it = ls.m_outs.begin(); it != ls.m_outs.end(); ++it)
    if (ls.m_curr_level.GetErrorLevel() <= it -> m_desired_level.GetErrorLevel())
      *it -> mp_out << ls.FormatMessage(ls.str(), it -> m_fl_add_time, it -> m_fl_add_level) << std::endl;

  ls.str(std::string());

  return os;
  }
//--------------------------------------------------------------------
