/*****************************************************************************
 *
 * Unrawel, Copyright (C) 2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

// UnrawelCmd.cpp : Defines the entry point for the console application.
//
#include "stdafx.h"

#include <LibUnrawel.h>

#include <iostream>
#include <cmath>
#include <algorithm>

#include "ParamsDispatcher.h"

#include "Processor.h"

#include "ImageTransfer.h"

#include "Magick.h"

#include <lcms.h>

//-----------------------------------------------------------------------------
#ifdef _MSC_VER
int _tmain(int argc, _TCHAR* argv[])
#else
int main(int argc, char* argv[])
#endif
  {
#ifndef USE_IMAGE_MAGICK
  std::cout << "Unrawel 1.03 alpha (cmd edition)" << std::endl;
  std::cout << "Compiled without ImageMagick support - some functions will be unavailable!" << std::endl;
#else
  std::cout << "Unrawel 1.03 alpha (cmd edition, with ImageMagick support)" << std::endl;
#endif
  std::cout << "Copyright (C) 2002-2003 by Alexander Tsvyashchenko <spam@ndl.kiev.ua>" << std::endl;
  std::cout << "This program comes with ABSOLUTELY NO WARRANTY; see copying.txt for details" << std::endl;
  std::cout << "Some algorithms have special licenses; see copyrights.txt for details" << std::endl;
  std::cout << "Unrawel contains work of other people; see acknowledgements.txt for details" << std::endl;
  std::cout << std::endl;

  if (argc < 4 || argc > 5)
    {
    std::cerr << "Usage: UnrawelCmd <config-file> <input-file> <output-file> [<log-file>]" << std::endl;
    return 1;
    }

  mg_log.SetOutput(std::cout);
  mg_log.SetLevel(INFO);
  mg_log.SetAddTime(false);
  mg_log.SetAddLevel(false);

  std::ofstream testlog;

  if (argc == 5)
    {
    testlog.open(argv[4]);

    mg_log.AddOutput(testlog);
    mg_log.SetLevel(DEBUG);
    mg_log.SetAddTime(true);
    mg_log.SetAddLevel(true);
    }
  
  try
    {
    mg_log << DEBUG << "sizeof(Pixel)=" << sizeof(Pixel) << endlog;

    if (sizeof(Pixel) != 6)
      mg_log << CRIT << "sizeof(Pixel) is not 6 - expect problems with image loading/saving" << endlog;

#ifdef USE_IMAGE_MAGICK
    mg_log << TRACE << "Initializing ImageMagick" << endlog;

    Magick::InitializeMagick(argv[0]);
#endif // #ifdef USE_IMAGE_MAGICK

    cmsErrorAction(LCMS_ERROR_SHOW);

    std::ifstream ifs(argv[1]);

    if (!ifs.is_open())
      {
      mg_log << CRIT << "Cannot open config file" << endlog;
      throw "Cannot open config file";
      }

    Processor::GetProcessor().SetFileToLoad(argv[2]);

    std::string str;

    for (std::getline(ifs, str); !ifs.eof(); std::getline(ifs, str))
      if (str[0] != '#')
	Processor::GetProcessor().Process(str);

    mg_log << INFO << "Writing result image to " << argv[3] << endlog;

    ImageTransfer::GetTransfer().Write(argv[3], Processor::GetProcessor().GetImage());

#ifdef USE_IMAGE_MAGICK
    mg_log << TRACE << "Finalizing ImageMagick" << endlog;

    Magick::DestroyMagick();
#endif // #ifdef USE_IMAGE_MAGICK
    }
  catch (MyException ex)
    {
    mg_log << CRIT << "Unknown exception was thrown by the program:"
      << std::endl << ex << endlog;
    mg_log.PrintAllRecentLogMessages();
    }
  catch(...)
    {
    mg_log << CRIT << "Unknown exception" << endlog;
    mg_log.PrintAllRecentLogMessages();
    }

  return 0;
  }
//-----------------------------------------------------------------------------

