/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef PARAMS_DEF_H
#define PARAMS_DEF_H

#include "MyException.h"

//-----------------------------------------------------------------------------
class IProcessor;
class Image;
class RawImageInfo;
//-----------------------------------------------------------------------------
struct IParams
  {
  public:
    virtual IParams* Clone() const = 0;
    virtual IParams& operator =(const IParams& iparams) = 0;
    virtual ~IParams() {}
  };
//-----------------------------------------------------------------------------
class IParamDef
  {
  public:
    IParamDef(const std::string& name,
      const std::string& description) { m_name = name;
                                        m_description = description; }

    virtual ~IParamDef() {}

  protected:
    std::string m_name, m_description;
  };
//-----------------------------------------------------------------------------
struct ParamLong: public IParamDef
  {
  ParamLong(const std::string& name, const std::string& description,
    long* p_val, long min_val = std::numeric_limits<long>::min(),
    long max_val = std::numeric_limits<long>::max()):
  IParamDef(name, description),
  mp_val(p_val),
  m_min_val(min_val),
  m_max_val(max_val)
    {
    m_def_val = *mp_val;
    }

  private:
    long m_def_val;
    long* mp_val;
    long m_min_val, m_max_val;
  };
//-----------------------------------------------------------------------------
struct ParamString: public IParamDef
  {
  ParamString(const std::string& name, const std::string& description,
              std::string* p_val): IParamDef(name, description),
                                   mp_val(p_val)
    {
    m_def_val = *mp_val;
    }

  private:
    std::string m_def_val;
    std::string* mp_val;
  };
//-----------------------------------------------------------------------------
struct ModuleDef
  {
  std::string m_module_decription;
  IParams* mp_params;
  std::list<IParamDef*> m_params_def;
  IProcessor* (*m_creator)();
  };
//-----------------------------------------------------------------------------
struct ModuleInfo
  {
  std::string m_name;
  std::list<IParams*> m_params;

  ModuleInfo() {}

  ModuleInfo(const ModuleInfo& info)
    {
    *this = info;
    }

  ModuleInfo& operator =(const ModuleInfo& info)
    {
    m_name = info.m_name;

    for (std::list<IParams*>::const_iterator it = info.m_params.begin();
         it != info.m_params.end(); ++it)
      m_params.push_back((*it) -> Clone());

    return *this;
    }

  ~ModuleInfo()
    {
    for (std::list<IParams*>::iterator it = m_params.begin();
         it != m_params.end(); ++it)
      delete *it;
    }
  };
//-----------------------------------------------------------------------------
typedef std::list<ModuleInfo> ProcessInfo;
//-----------------------------------------------------------------------------
typedef std::map<std::string, ModuleDef> DefTable;
//-----------------------------------------------------------------------------
class DefTableSingleton
  {
  public:
    static DefTable& GetDefTable();

  private:
    static bool m_fl_initialized;

  private:
    static void FillDefTable(DefTable& table);
  };
//-----------------------------------------------------------------------------
class IProcessor
  {
  public:
    class ExIProcessorError: public MyException
      { 
      public:
        ExIProcessorError(const std::string& i_err_id, const std::string& msg): MyException("IProcessor", i_err_id, msg) {}
      };

  public:
    virtual void Process(Image& image, RawImageInfo& image_info, const std::string& method) = 0;
    virtual ~IProcessor() {}
    virtual void GetParameters(IParams& o_params) const = 0;
    virtual void SetParameters(IParams& i_params) = 0;
  };
//-----------------------------------------------------------------------------

#endif // PARAMS_DEF_H
